@@ -82,83 +82,85 @@ static char *auxdata;
 static int auxlen;
 static int showflag = 0,showflag2 = 0;
 
+#define NULLBINCMD ((int (*) DCLPROTO((char *,char **,char *,int))) 0)
+
 struct bincmd builtins[] = {
-	"[",bin_test,0,-1,0,BIN_BRACKET,NULL,NULL,
-	".",bin_dot,1,-1,0,0,NULL,NULL,
-	":",bin_colon,0,-1,0,0,NULL,NULL,
-	"alias",bin_alias,0,-1,0,0,"ga",NULL,
-	"autoload",bin_typeset,0,-1,BINF_TYPEOPTS,0,"tx","fu",
-	"bg",bin_fg,0,-1,0,BIN_BG,NULL,NULL,
-	"bindkey",bin_bindkey,0,-1,0,0,"asvemdrl",NULL,
-	"break",bin_break,0,1,0,BIN_BREAK,NULL,NULL,
-	"builtin",NULL,0,0,0,BIN_BUILTIN,NULL,NULL,
-	"bye",bin_break,0,1,0,BIN_EXIT,NULL,NULL,
-	"cd",bin_cd,0,2,0,BIN_CD,NULL,NULL,
-	"chdir",bin_cd,0,2,0,BIN_CD,NULL,NULL,
-	"compctl",bin_compctl,0,-1,0,0,NULL,NULL,
-	"continue",bin_break,0,1,0,BIN_CONTINUE,NULL,NULL,
-	"declare",bin_typeset,0,-1,BINF_TYPEOPTS,0,"LRZfilrtux",NULL,
-	"dirs",bin_dirs,0,-1,0,0,"v",NULL,
-	"disable",bin_disable,1,-1,0,0,NULL,NULL,
-	"disown",bin_fg,1,-1,0,BIN_DISOWN,NULL,NULL,
-	"echo",bin_print,0,-1,BINF_PRINTOPTS|BINF_ECHOPTS,BIN_PRINT,"n","-",
-	"echotc",bin_echotc,1,-1,0,0,NULL,NULL,
-	"enable",bin_enable,1,-1,0,0,NULL,NULL,
-	"eval",bin_eval,0,-1,0,BIN_EVAL,NULL,NULL,
-	"exit",bin_break,0,1,0,BIN_EXIT,NULL,NULL,
-	"export",bin_typeset,0,-1,BINF_TYPEOPTS,0,"LRZfilrtu","x",
-	"false",bin_let,0,0,0,0,NULL,NULL,
-	"fc",bin_fc,0,-1,BINF_FCOPTS,BIN_FC,"nlreRWAdD",NULL,
-	"fg",bin_fg,0,-1,0,BIN_FG,NULL,NULL,
-	"functions",bin_typeset,0,-1,BINF_TYPEOPTS,0,"tu","f",
-	"getln",bin_read,0,-1,0,0,NULL,"zr",
-	"getopts",bin_getopts,2,-1,0,0,NULL,NULL,
-	"hash",bin_hash,2,2,0,0,"r",NULL,
-	"history",bin_fc,0,-1,0,BIN_FC,"nrdD","l",
-	"integer",bin_typeset,0,-1,BINF_TYPEOPTS,0,"LRZlrtux","i",
-	"jobs",bin_fg,0,-1,0,BIN_JOBS,"lpZ",NULL,
-	"kill",bin_kill,0,-1,0,0,NULL,NULL,
-	"let",bin_let,1,-1,0,0,NULL,NULL,
-	"limit",bin_limit,0,-1,0,0,"sh",NULL,
-	"local",bin_typeset,0,-1,BINF_TYPEOPTS,0,"LRZfilrtux",NULL,
-	"log",bin_log,0,0,0,0,NULL,NULL,
-	"logout",bin_break,0,1,0,BIN_LOGOUT,NULL,NULL,
-	"popd",bin_cd,0,2,0,BIN_POPD,NULL,NULL,
-	"print",bin_print,0,-1,BINF_PRINTOPTS,BIN_PRINT,"RDPnrslzNu0123456789p-",NULL,
-	"pushd",bin_cd,0,2,0,BIN_PUSHD,NULL,NULL,
-	"pushln",bin_print,0,-1,BINF_PRINTOPTS,BIN_PRINT,NULL,"-nz",
-	"pwd",bin_pwd,0,0,0,0,NULL,NULL,
-	"r",bin_fc,0,-1,BINF_R,BIN_FC,"nrl",NULL,
-	"read",bin_read,0,-1,0,0,"rzu0123456789p",NULL,
-	"readonly",bin_typeset,0,-1,BINF_TYPEOPTS,0,"LRZfiltux","r",
-	"rehash",bin_rehash,0,0,0,0,"f",NULL,
-	"return",bin_break,0,1,0,BIN_RETURN,NULL,NULL,
-	"sched",bin_sched,0,-1,0,0,NULL,NULL,
-	"set",bin_set,0,-1,BINF_SETOPTS|BINF_PLUSOPTS,0,"0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZaefghijklnosuvwxy",NULL,
-	"setopt",bin_setopt,0,-1,BINF_PLUSOPTS,0,"0123456789BCDEFGHIJKLMNOPQRSTUVWXYZaefghijklmnosuvwxy",NULL,
-	"shift",bin_break,0,1,0,BIN_SHIFT,NULL,NULL,
-	"source",bin_dot,1,-1,0,0,NULL,NULL,
-	"suspend",bin_suspend,0,0,0,0,"f",NULL,
-	"test",bin_test,0,-1,0,BIN_TEST,NULL,NULL,
-	"ttyctl",bin_ttyctl,0,0,0,0,"fu",NULL,
-	"times",bin_times,0,0,0,0,NULL,NULL,
-	"trap",bin_trap,0,-1,0,0,NULL,NULL,
-	"true",bin_colon,0,0,0,0,NULL,NULL,
-	"type",bin_whence,0,-1,0,0,"pfa","v",
-	"typeset",bin_typeset,0,-1,BINF_TYPEOPTS,0,"LRZfilrtux",NULL,
-	"ulimit",bin_ulimit,0,1,0,0,"HSacdfmnt",NULL,
-	"umask",bin_umask,0,1,0,0,NULL,NULL,
-	"unalias",bin_unalias,1,-1,0,0,NULL,NULL,
-	"unfunction",bin_unhash,1,-1,0,0,NULL,NULL,
-	"unhash",bin_unhash,1,-1,0,0,NULL,NULL,
-	"unlimit",bin_unlimit,0,-1,0,0,"h",NULL,
-	"unset",bin_unset,1,-1,0,0,NULL,NULL,
-	"unsetopt",bin_setopt,0,-1,BINF_PLUSOPTS,1,"0123456789BCDEFGHIJKLMNOPQRSTUWXYZabefghijklmnosuvwxy",NULL,
-	"vared",bin_vared,1,1,0,0,NULL,NULL,
-	"wait",bin_fg,0,-1,0,BIN_WAIT,NULL,NULL,
-	"whence",bin_whence,0,-1,0,0,"pvcfa",NULL,
-	"which",bin_whence,0,-1,0,0,"pa","c",
-	NULL,NULL,0,0,0,0,NULL,NULL
+	{"[",bin_test,0,-1,0,BIN_BRACKET,NULL,NULL},
+	{".",bin_dot,1,-1,0,0,NULL,NULL},
+	{":",bin_colon,0,-1,0,0,NULL,NULL},
+	{"alias",bin_alias,0,-1,0,0,"ga",NULL},
+	{"autoload",bin_typeset,0,-1,BINF_TYPEOPTS,0,"tx","fu"},
+	{"bg",bin_fg,0,-1,0,BIN_BG,NULL,NULL},
+	{"bindkey",bin_bindkey,0,-1,0,0,"asvemdrl",NULL},
+	{"break",bin_break,0,1,0,BIN_BREAK,NULL,NULL},
+	{"builtin",NULLBINCMD,0,0,0,BIN_BUILTIN,NULL,NULL},
+	{"bye",bin_break,0,1,0,BIN_EXIT,NULL,NULL},
+	{"cd",bin_cd,0,2,0,BIN_CD,NULL,NULL},
+	{"chdir",bin_cd,0,2,0,BIN_CD,NULL,NULL},
+	{"compctl",bin_compctl,0,-1,0,0,NULL,NULL},
+	{"continue",bin_break,0,1,0,BIN_CONTINUE,NULL,NULL},
+	{"declare",bin_typeset,0,-1,BINF_TYPEOPTS,0,"LRZfilrtux",NULL},
+	{"dirs",bin_dirs,0,-1,0,0,"v",NULL},
+	{"disable",bin_disable,1,-1,0,0,NULL,NULL},
+	{"disown",bin_fg,1,-1,0,BIN_DISOWN,NULL,NULL},
+	{"echo",bin_print,0,-1,BINF_PRINTOPTS|BINF_ECHOPTS,BIN_PRINT,"n","-"},
+	{"echotc",bin_echotc,1,-1,0,0,NULL,NULL},
+	{"enable",bin_enable,1,-1,0,0,NULL,NULL},
+	{"eval",bin_eval,0,-1,0,BIN_EVAL,NULL,NULL},
+	{"exit",bin_break,0,1,0,BIN_EXIT,NULL,NULL},
+	{"export",bin_typeset,0,-1,BINF_TYPEOPTS,0,"LRZfilrtu","x"},
+	{"false",bin_let,0,0,0,0,NULL,NULL},
+	{"fc",bin_fc,0,-1,BINF_FCOPTS,BIN_FC,"nlreIRWAdDfE",NULL},
+	{"fg",bin_fg,0,-1,0,BIN_FG,NULL,NULL},
+	{"functions",bin_typeset,0,-1,BINF_TYPEOPTS,0,"tu","f"},
+	{"getln",bin_read,0,-1,0,0,NULL,"zr"},
+	{"getopts",bin_getopts,2,-1,0,0,NULL,NULL},
+	{"hash",bin_hash,2,2,0,0,"r",NULL},
+	{"history",bin_fc,0,-1,0,BIN_FC,"nrdDfE","l"},
+	{"integer",bin_typeset,0,-1,BINF_TYPEOPTS,0,"LRZlrtux","i"},
+	{"jobs",bin_fg,0,-1,0,BIN_JOBS,"lpZ",NULL},
+	{"kill",bin_kill,0,-1,0,0,NULL,NULL},
+	{"let",bin_let,1,-1,0,0,NULL,NULL},
+	{"limit",bin_limit,0,-1,0,0,"sh",NULL},
+	{"local",bin_typeset,0,-1,BINF_TYPEOPTS,0,"LRZfilrtux",NULL},
+	{"log",bin_log,0,0,0,0,NULL,NULL},
+	{"logout",bin_break,0,1,0,BIN_LOGOUT,NULL,NULL},
+	{"popd",bin_cd,0,2,0,BIN_POPD,NULL,NULL},
+	{"print",bin_print,0,-1,BINF_PRINTOPTS,BIN_PRINT,"RDPnrslzNu0123456789p-",NULL},
+	{"pushd",bin_cd,0,2,0,BIN_PUSHD,NULL,NULL},
+	{"pushln",bin_print,0,-1,BINF_PRINTOPTS,BIN_PRINT,NULL,"-nz"},
+	{"pwd",bin_pwd,0,0,0,0,NULL,NULL},
+	{"r",bin_fc,0,-1,BINF_R,BIN_FC,"nrl",NULL},
+	{"read",bin_read,0,-1,0,0,"rzu0123456789p",NULL},
+	{"readonly",bin_typeset,0,-1,BINF_TYPEOPTS,0,"LRZfiltux","r"},
+	{"rehash",bin_rehash,0,0,0,0,"f",NULL},
+	{"return",bin_break,0,1,0,BIN_RETURN,NULL,NULL},
+	{"sched",bin_sched,0,-1,0,0,NULL,NULL},
+	{"set",bin_set,0,-1,BINF_SETOPTS|BINF_PLUSOPTS,0,"0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZaefghjklmnosuvwxy",NULL},
+	{"setopt",bin_setopt,0,-1,BINF_PLUSOPTS,0,"0123456789BCDEFGHIJKLMNOPQRSTUVWXYZaefghjklmnosuvwxy",NULL},
+	{"shift",bin_break,0,1,0,BIN_SHIFT,NULL,NULL},
+	{"source",bin_dot,1,-1,0,0,NULL,NULL},
+	{"suspend",bin_suspend,0,0,0,0,"f",NULL},
+	{"test",bin_test,0,-1,0,BIN_TEST,NULL,NULL},
+	{"ttyctl",bin_ttyctl,0,0,0,0,"fu",NULL},
+	{"times",bin_times,0,0,0,0,NULL,NULL},
+	{"trap",bin_trap,0,-1,0,0,NULL,NULL},
+	{"true",bin_colon,0,0,0,0,NULL,NULL},
+	{"type",bin_whence,0,-1,0,0,"pfa","v"},
+	{"typeset",bin_typeset,0,-1,BINF_TYPEOPTS,0,"LRZfilrtux",NULL},
+	{"ulimit",bin_ulimit,0,1,0,0,"HSacdfmnt",NULL},
+	{"umask",bin_umask,0,1,0,0,NULL,NULL},
+	{"unalias",bin_unalias,1,-1,0,0,NULL,NULL},
+	{"unfunction",bin_unhash,1,-1,0,0,NULL,NULL},
+	{"unhash",bin_unhash,1,-1,0,0,NULL,NULL},
+	{"unlimit",bin_unlimit,0,-1,0,0,"h",NULL},
+	{"unset",bin_unset,1,-1,0,0,NULL,NULL},
+	{"unsetopt",bin_setopt,0,-1,BINF_PLUSOPTS,1,"0123456789BCDEFGHIJKLMNOPQRSTUWXYZabefghjklmnosuvwxy",NULL},
+	{"vared",bin_vared,1,1,0,0,NULL,NULL},
+	{"wait",bin_fg,0,-1,0,BIN_WAIT,NULL,NULL},
+	{"whence",bin_whence,0,-1,0,0,"pvcfa",NULL},
+	{"which",bin_whence,0,-1,0,0,"pa","c"},
+	{NULL,NULLBINCMD,0,0,0,0,NULL,NULL}
 	};
 
 /* print options */
@@ -170,11 +172,10 @@ struct option *opp;
 	if (isset(KSHOPTIONPRINT)) {
 		printf("Current option settings\n");
 		for (opp = optns; opp->name; opp++)
-			printf("%-20s%s\n", opp->name,
-				(opts[opp->id] == OPT_SET) ? "on" : "off");
+			printf("%-20s%s\n", opp->name, isset(opp->id) ? "on" : "off");
 	} else
 		for (opp = optns; opp->name; opp++)
-			if (opts[opp->id] == OPT_SET)
+			if (isset(opp->id))
 				puts(opp->name);
 }
 
@@ -267,6 +268,7 @@ int num = -1;
 			break;
 		case BIN_RETURN:
 			retflag = 1;
+			breaks = loops;
 			return lastval = (num == -1) ? lastval : num;
 		case BIN_SHIFT:
 			{
@@ -302,10 +304,10 @@ int job,lng,firstjob = -1,retval = 0;
 		return 1;
 		}
 	if (unset(NOTIFY)) scanjobs();
-	if (!(jobtab[curjob].stat & STAT_INUSE))
+	if (curjob != -1 && !(jobtab[curjob].stat & STAT_INUSE))
 		{
 		curjob = prevjob; setprevjob();
-		if (!(jobtab[curjob].stat & STAT_INUSE))
+		if (curjob != -1 && !(jobtab[curjob].stat & STAT_INUSE))
 			curjob = prevjob; setprevjob();
 		}
 	if (func == BIN_JOBS)
@@ -456,9 +458,15 @@ int val;
 				return 1;
 			}
 		} else killrunjobs();
-	savehistfile(getsparam("HISTFILE"),0,0);
-	if (islogin && unset(NORCS))
-		sourcehome(".zlogout");
+	if (unset(NORCS) && interact) {
+		if (isset(APPENDHISTORY)) {
+			savehistfile(getsparam("HISTFILE"),0,3);
+			readhistfile(getsparam("HISTFILE"),0);
+		}
+		savehistfile(getsparam("HISTFILE"),0,0);
+		if (islogin)
+			sourcehome(".zlogout");
+	}
 	if (sigtrapped[SIGEXIT])
 		dotrap(SIGEXIT);
 	exit(val); return 0;
@@ -501,14 +509,14 @@ int c;
 	}
 	for (opp = optns; opp->name; opp++)
 		if (ops[opp->id] == 1+isun)
-			opts[opp->id] = OPT_SET;
-		else if (ops[opp->id] == 2-isun)
-			opts[opp->id] = OPT_UNSET;
+			opts[(int)opp->id] = OPT_SET;
+		else if (ops[(int)opp->id] == 2-isun)
+			opts[(int)opp->id] = OPT_UNSET;
 	while (*args) {
 		c = optlookup(*args++);
 		if (c != -1) {
-			if (c == INTERACTIVE || c == MONITOR)
-				zerrnam(nam,"can't change that option",NULL,0);
+			if (c == INTERACTIVE)
+				zerrnam(nam,"can't change option: %s",args[-1],0);
 			else
 				opts[c] = (isun) ? OPT_UNSET : OPT_SET;
 		} else {
@@ -521,15 +529,47 @@ int c;
 
 /* execute func on each member of the hash table ht */
 
+static hnamcmp DCLPROTO((struct hashnode **a,struct hashnode **b));
+static int hnamcmp(a,b)
+struct hashnode **a;struct hashnode **b;
+{
+	return forstrcmp(&((*a)->nam), &((*b)->nam));
+}
+
 void listhtable(ht,func) /**/
 Hashtab ht;HFunc func;
 {
 int t0;
 struct hashnode *hn;
 
+#ifndef HASHORDER
+
+int nhash; struct hashnode **hnsorttab, **htp;
+
+	for (nhash = 0, t0 = ht->hsize-1; t0 >= 0; t0--)
+		for (hn = ht->nodes[t0]; hn; hn = hn->next)
+			nhash++;
+
+	hnsorttab = (struct hashnode **)zalloc (nhash * sizeof(struct hashnode *));
+	for (htp = hnsorttab, t0 = ht->hsize-1; t0 >= 0; t0--)
+		for (hn = ht->nodes[t0]; hn; hn = hn->next)
+			*htp++ = hn;
+
+	qsort((vptr)&hnsorttab[0],nhash,sizeof(struct hashnode *),
+			(int (*) DCLPROTO((const void *, const void *)))hnamcmp);
+
+	for (htp = hnsorttab; nhash; nhash--, htp++)
+		func((*htp)->nam, (char *) (*htp));
+
+	free (hnsorttab);
+
+#else
+
 	for (t0 = ht->hsize-1; t0 >= 0; t0--)
 		for (hn = ht->nodes[t0]; hn; hn = hn->next)
 			func(hn->nam,(char *) hn);
+
+#endif HASHORDER
 }
 
 /* print a shell function (used with listhtable) */
@@ -666,8 +706,10 @@ int retval = 0;
 int csh = ops['c'],all = ops['a'];
 int v = ops['v'] || csh;
 char *cnam;
+int informed;
 
 	for (; *argv; argv++) {
+		informed = 0;
 		if (!ops['p'] && (a = (Alias) gethnode(*argv,aliastab))) {
 			if (a->cmd < 0)
 				printf((csh) ? "%s: shell reserved word\n" :
@@ -682,6 +724,7 @@ char *cnam;
 					"%s is a global alias for %s\n",*argv,a->text);
 			retval = 0;
 			if (!all) continue;
+			informed = 1;
 		}
 		if (!ops['p'] && (chn = (Cmdnam) gethnode(*argv,cmdnamtab)) &&
 				(chn->type == SHFUNC || chn->type == BUILTIN)) {
@@ -697,6 +740,7 @@ char *cnam;
 					(v) ? "%s is a shell builtin\n" : "%s\n",*argv);
 			retval = 0;
 			if (!all) continue;
+			informed = 1;
 		}
 		if (all) {
 			char **pp,buf[MAXPATHLEN],*z;
@@ -709,8 +753,14 @@ char *cnam;
 					if (v && !csh) printf("%s is %s\n",*argv,buf);
 					else puts(buf);
 					retval = 0;
+					informed = 1;
 				}
 			}
+			if (!informed && v) {
+				printf("%s not found\n",*argv);
+				retval = 1;
+				break;
+			}
 		} else if (!(cnam = findcmd(*argv))) {
 			if (v) printf("%s not found\n",*argv);
 			retval = 1;
@@ -719,6 +769,7 @@ char *cnam;
 			if (v && !csh) printf("%s is %s\n",*argv,cnam);
 			else puts(cnam);
 			retval = 0;
+			free(cnam);
 		}
 	}
 	return retval;
@@ -729,15 +780,20 @@ char *cnam;
 int bin_cd(nam,argv,ops,func) /**/
 char *nam;char **argv;char *ops;int func;
 {
-char *dest;
+char *dest, *dir;
 
 	if (func == BIN_CD && isset(AUTOPUSHD))
 		func = BIN_PUSHD;
-	dest = cd_get_dest(nam,argv,ops,func);
-	if (!dest) return 1;
+	dir = dest = cd_get_dest(nam,argv,ops,func);
+	if (!dest)
+		return 1;
 	dest = cd_do_chdir(nam,dest);
-	if (!dest) return 1;
+	if (dest != dir)
+		free(dir);
+	if (!dest)
+		return 1;
 	cd_new_pwd(func,dest);
+	free(dest);
 	return 0;
 }
 
@@ -747,19 +803,20 @@ char *nam;char **argv;char *ops;int func;
 char *dest;
 
 	if (!argv[0])
-		if (func == BIN_CD || (func == BIN_PUSHD && isset(PUSHDTOHOME)
-				|| !full(dirstack)))
-			dest = home;
+		if (func == BIN_CD || ((func == BIN_PUSHD && isset(PUSHDTOHOME))
+				|| empty(dirstack)))
+			dest = ztrdup(home);
 		else
-			dest = getnode(dirstack);
+			dest = ztrdup(getnode(dirstack));
 	else if (!argv[1]) {
 		Lknode n;
 		int dd;
+		char *end;
 
 		if (argv[0][1] && argv[0][0] == (isset(PUSHDMINUS) ? '-' : '+')) {
-			dd = atoi(argv[0]+1)-1;
-			if (dd < 0) {
-				zerrnam(nam,"bad directory specification",NULL,0);
+			dd = zstrtol(argv[0]+1,&end,10)-1;
+			if (dd < 0 || *end != '\0') {
+				zerrnam(nam,"%s: bad directory specification",argv[0],0);
 				return NULL;
 			}
 			for (n = firstnode(dirstack); n && dd; dd--, incnode(n));
@@ -769,7 +826,11 @@ char *dest;
 			}
 			dest = remnode(dirstack,n);
 		} else if (argv[0][1] && argv[0][0] == (isset(PUSHDMINUS) ? '+' : '-')) {
-			dd = atoi(argv[0]+1);
+			dd = zstrtol(argv[0]+1,&end,10);
+			if (*end != '\0') {
+				zerrnam(nam,"%s: bad directory specification",argv[0],0);
+				return NULL;
+			}
 			for (n = lastnode(dirstack); n != (Lknode) dirstack && dd;
 					dd--, n = prevnode(n));
 			if (n == (Lknode) dirstack) {
@@ -778,8 +839,8 @@ char *dest;
 			}
 			dest = remnode(dirstack,n);
 		} else {
-			if (!strcmp(argv[0],"-")) printdircr(dest = oldpwd);
-			else dest = argv[0];
+			if (!strcmp(argv[0],"-")) printdircr(dest = ztrdup(oldpwd));
+			else dest = ztrdup(argv[0]);
 		}
 	} else {
 		char *u;
@@ -792,7 +853,7 @@ char *dest;
 		len1 = strlen(argv[0]);
 		len2 = strlen(argv[1]);
 		len3 = u-pwd;
-		dest = alloc(len3+len2+strlen(u+len1)+1);
+		dest = zalloc(len3+len2+strlen(u+len1)+1);
 		strncpy(dest,pwd,len3);
 		strcpy(dest+len3,argv[1]);
 		strcat(dest,u+len1);
@@ -831,7 +892,7 @@ char **pp,*ret;
 		char *s = getsparam(dest);
 		if (s && *s == '/' && chdir(s) != -1) {
 			printdircr(s);
-			return s;
+			return ztrdup(s);
 		}
 		if (errno != ENOENT) eno = errno;
 	}
@@ -842,26 +903,26 @@ char **pp,*ret;
 char *cd_try_chdir(pfix,dest) /**/
 char *pfix; char *dest;
 {
-static char buf[MAXPATHLEN], buf2[MAXPATHLEN];
+char buf[MAXPATHLEN], buf2[MAXPATHLEN];
 char *s;
 int dotsct;
 
 	if (pfix) sprintf(buf,"%s/%s",(!strcmp("/",pfix)) ? "" : pfix,dest); 
 	else strcpy(buf,dest);
 	dotsct = fixdir(buf2,buf);
-	if (buf2[0] == '/') return (chdir(buf2) == -1) ? NULL : buf2;
+	if (buf2[0] == '/') return (chdir(buf) == -1) ? NULL : ztrdup(buf2);
 	if (!dotsct) {
-		if (chdir((*buf2) ? buf2 : ".") == -1) return NULL;
+		if (chdir(buf) == -1) return NULL;
 		if (*buf2) sprintf(buf,"%s/%s",(!strcmp("/",pwd)) ? "" : pwd,buf2);
 		else strcpy(buf,pwd);
-		return buf;
+		return ztrdup(buf);
 	}
 	strcpy(buf,pwd);
 	s = buf+strlen(buf)-1;
 	while (dotsct--) while (s != buf) if (*--s == '/') break;
 	if (s == buf || *buf2) s++;
 	strcpy(s,buf2);
-	if (chdir(buf) != -1 || chdir(dest) != -1) return buf;
+	if (chdir(buf) != -1 || chdir(dest) != -1) return ztrdup(buf);
 	return NULL;
 }
 
@@ -872,11 +933,14 @@ int ct = 0;
 char *d0 = d;
 
 #ifdef HAS_RFS
-	if (*s == '/' && s[1] == '.' && s[2] == '.') {
+	while (*s == '/' && s[1] == '.' && s[2] == '.') {
 		*d++ = '/'; *d++ = '.'; *d++ = '.';
 		s += 3;
 	}
 #endif
+#ifdef apollo
+	if (*s == '/') *d++ = *s++;   /*added RBC 18/05/92 */
+#endif	
 	for (;;) {
 		if (*s == '/') {
 			*d++ = *s++;
@@ -890,6 +954,7 @@ char *d0 = d;
 		if (s[0] == '.' && s[1] == '.' && (s[2] == '\0' || s[2] == '/')) {
 			if (d > d0+1) {
 				for (d--; d > d0+1 && d[-1] != '/'; d--);
+				if (d[-1] != '/') d--;
 			} else ct++;
 			s += 2; if (*s) s++;
 		} else if (s[0] == '.' && (s[1] == '/' || s[1] == '\0')) {
@@ -905,12 +970,27 @@ int func; char *s;
 {
 Param pm;
 List l;
+char *new_pwd;
 
-	oldpwd = pwd;
 	if (isset(CHASELINKS))
-		pwd = findpwd(s);
+		new_pwd = findpwd(s);
 	else
-		pwd = ztrdup(s);
+		new_pwd = ztrdup(s);
+	if (!strcmp(new_pwd, pwd)) {
+		free(new_pwd);
+#ifdef ALWAYS_DO_CD_PROCESSING
+		if (unset(PUSHDSILENT) && func != BIN_CD && isset(INTERACTIVE))
+			pdstack();
+		if (l = getshfunc("chpwd")) {
+			fflush(stdout); fflush(stderr);
+			doshfuncnoval(dupstruct(l),NULL,0);
+		}
+#endif
+		return;
+	}
+	free(oldpwd);
+	oldpwd = pwd;
+	pwd = new_pwd;
 	if ((pm = gethnode("PWD", paramtab)) &&
 		 (pm->flags & PMFLAG_x) && pm->env)
 		pm->env = replenv(pm->env,pwd);
@@ -921,12 +1001,15 @@ List l;
 		permalloc();
 		if (isset(PUSHDIGNOREDUPS)) {
 			Lknode n;
-			for (n = firstnode(dirstack); n; incnode(n))
-				if (!strcmp(oldpwd,getdata(n))) {
+			char *nodedata;
+			for (n = firstnode(dirstack); n; incnode(n)) {
+				nodedata = getdata(n);
+				if (!strcmp(oldpwd,nodedata) || !strcmp(pwd,nodedata)) {
 					free(remnode(dirstack,n)); break;
 				}
+			}	
 		}
-		pushnode(dirstack,oldpwd);
+		pushnode(dirstack,ztrdup(oldpwd));
 		heapalloc();
 	}
 	if (unset(PUSHDSILENT) && func != BIN_CD && isset(INTERACTIVE))
@@ -1088,14 +1171,14 @@ int retval = 0;
 		else {
 			if ((*argv)[1] == 'l' && (*argv)[2] == '\0') {
 				printf("%s",sigs[1]);
-				for (sig = 2; sig != SIGCOUNT; sig++)
+				for (sig = 2; sig <= SIGCOUNT; sig++)
 					printf(" %s",sigs[sig]);
 				putchar('\n');
 				return 0;
 			}
-			for (sig = 0; sig != SIGCOUNT; sig++)
+			for (sig = 1; sig <= SIGCOUNT; sig++)
 				if (!strcmp(sigs[sig],*argv+1)) break;
-			if (sig == SIGCOUNT) {
+			if (sig > SIGCOUNT) {
 				zerrnam(nam,"unknown signal: SIG%s",*argv+1,0);
 				zerrnam(nam,"type kill -l for a List of signals",NULL,0);
 				return 1;
@@ -1125,17 +1208,39 @@ int retval = 0;
 			}
 		} else if (!isanum(*argv)) {
 			zerrnam("kill","illegal pid: %s",*argv,0);
+			retval = 1;
 		} else if (kill(atoi(*argv),sig) == -1) {
-				zerrnam("kill","kill failed: %e",NULL,errno);
-				retval = 1;
+			zerrnam("kill","kill failed: %e",NULL,errno);
+			retval = 1;
 		}
 	}
-	return 0;
+	return retval;
 }
 
 static char *recs[] = {
 	"cputime","filesize","datasize","stacksize","coredumpsize",
-	"resident","descriptors"
+#ifdef RLIMIT_RSS
+#ifdef RLIMIT_MEMLOCK
+	"memoryuse",
+#else
+	"resident",
+#endif /* RLIMIT_MEMLOCK */
+#endif /* RLIMIT_RSS */
+#ifdef RLIMIT_MEMLOCK
+	"memorylocked",
+#endif
+#ifdef RLIMIT_NPROC
+	"maxproc",
+#endif
+#ifdef RLIMIT_OFILE
+	"openfiles",
+#endif
+#ifdef RLIMIT_NOFILE
+	"descriptors",
+#endif
+#ifdef RLIMIT_VMEM
+	"memorysize"
+#endif
 	};
 
 int bin_limit(nam,argv,ops,func) /**/
@@ -1201,6 +1306,14 @@ long val;
 					return 1;
 					}
 			}
+#ifdef RLIMIT_NPROC
+		else if (lim == RLIMIT_NPROC)
+			val = zstrtol(s,&s,10);
+#endif
+#ifdef RLIMIT_OFILE
+		else if (lim == RLIMIT_OFILE)
+			val = zstrtol(s,&s,10);
+#endif
 #ifdef RLIMIT_NOFILE
 		else if (lim == RLIMIT_NOFILE)
 			val = zstrtol(s,&s,10);
@@ -1311,6 +1424,14 @@ long val;
 			else if (!t0)
 				printf("%d:%02d:%02d\n",(int) (val/3600),
 					(int) (val/60) % 60,(int) (val % 60));
+#ifdef RLIMIT_NPROC
+			else if (t0 == RLIMIT_NPROC)
+				printf("%d\n",(int) val);
+#endif
+#ifdef RLIMIT_OFILE
+			else if (t0 == RLIMIT_OFILE)
+				printf("%d\n",(int) val);
+#endif
 #ifdef RLIMIT_NOFILE
 			else if (t0 == RLIMIT_NOFILE)
 				printf("%d\n",(int) val);
@@ -1497,6 +1618,11 @@ int done = 0;
 char *s,*hs;
 Histent ent;
 
+	if (r) {
+		r = last;
+		last = first;
+		first = r;
+	}
 	if (!subs) done = 1;
 	for (;;) {
 		hs = quietgetevent(first);
@@ -1513,7 +1639,19 @@ Histent ent;
 			
 			if (!ent) ent = gethistent(first);
 			ltm = localtime(&ent->stim);
-			fprintf(f,"%2d:%02d  ",ltm->tm_hour,ltm->tm_min);
+			if (d >= 2) {
+				if (d >= 4) {
+					fprintf(f,"%d.%d.%d ",
+						ltm->tm_mday,ltm->tm_mon+1,
+						ltm->tm_year+1900);
+				} else {
+					fprintf(f,"%d/%d/%d ",
+						ltm->tm_mon+1,
+						ltm->tm_mday,
+						ltm->tm_year+1900);
+		}
+			}
+			fprintf(f,"%02d:%02d  ",ltm->tm_hour, ltm->tm_min);
 		}
 		if (D) {
 			long diff;
@@ -1527,7 +1665,8 @@ Histent ent;
 			putc('\n',f);
 		} else fprintf(f,"%s\n",s);
 		if (first == last) break;
-		(r) ? first-- : first++;
+		else if (first > last) first--;
+		else first++;
 	}
 	if (f != stdout) fclose(f);
 	if (!done) {
@@ -1550,7 +1689,7 @@ char *ename;char *fn;
 int bin_fc(nam,argv,ops,func) /**/
 char *nam;char **argv;char *ops;int func;
 {
-int first = -1,last = -1,retval,minflag = 0;
+int first = -1,last = -1,retval,delayrem,minflag = 0;
 char *s;
 struct asgment *asgf = NULL,*asgl = NULL;
 
@@ -1558,17 +1697,22 @@ struct asgment *asgf = NULL,*asgl = NULL;
 		zerrnam(nam,"not interactive shell",NULL,0);
 		return 1;
 	}
-	if (!(ops['l'] && unset(HISTNOSTORE))) remhist();
+	delayrem = !strcmp(nam,"r")
+	      && argv[0]
+	      &&(argv[1] || !strchr(argv[0],'='));
+	if (!delayrem && !(ops['l'] && unset(HISTNOSTORE))) remhist();
 	if (ops['R']) {
 		readhistfile(*argv ? *argv : getsparam("HISTFILE"),1);
 		return 0;
 	}
 	if (ops['W']) {
-		savehistfile(*argv ? *argv : getsparam("HISTFILE"),1,0);
+		savehistfile(*argv ? *argv : getsparam("HISTFILE"),1,
+			     (ops['I'] ? 2 : 0));
 		return 0;
 	}
 	if (ops['A']) {
-		savehistfile(*argv ? *argv : getsparam("HISTFILE"),1,1);
+		savehistfile(*argv ? *argv : getsparam("HISTFILE"),1,
+			     (ops['I'] ? 3 : 1));
 		return 0;
 	}
 	while (*argv && equalsplit(*argv,&s)) {
@@ -1598,22 +1742,24 @@ struct asgment *asgf = NULL,*asgl = NULL;
 		zerrnam("fc","too many arguments",NULL,0);
 		return 1;
 	}
+	if (delayrem) remhist();
 	if (first == -1) first = (ops['l']) ? curhist-16 : curhist;
 	if (last == -1) last = (ops['l']) ? curhist : first;
 	if (first < firsthist()) first = firsthist();
 	if (last == -1) last = (minflag) ? curhist : first;
 	if (ops['l'])
-		retval = fclist(stdout,!ops['n'],ops['r'],ops['D'],ops['d'],
+		retval = fclist(stdout,!ops['n'],ops['r'],ops['D'],
+				ops['d'] + ops['f'] * 2 + ops['E'] * 4,
 			first,last,asgf);
 	else {
 		FILE *out;
 		char *fil = gettemp();
 
+		retval = 1;
 		out = fopen(fil,"w");
 		if (!out)
 			zerrnam("fc","can't open temp file: %e",NULL,errno);
 		else {
-			retval = 1;
 			if (!fclist(out,0,ops['r'],0,0,first,last,asgf))
 				if (fcedit(auxdata ? auxdata : fceditparam,fil))
 					if (stuff(fil))
@@ -1888,6 +2034,7 @@ FILE *fout = stdout;
 		ent->lex = ztrdup(join(args,HISTSPACE));
 		ent->lit = ztrdup(join(args,' '));
 		ent->stim = ent->ftim = time(NULL);
+		ent->flags = 0;
 		heapalloc();
 		return 0;
 	}
@@ -1912,7 +2059,7 @@ FILE *fout = stdout;
 		if (ops['D']) fprintdir(*args,fout);
 		else if (ops['P']) {
 			int junk;
-			fputs(putprompt(*args,&junk),fout);
+			fputs(putprompt(*args,&junk,0),fout);
 		} else fputs(*args,fout);
 		if (args[1]) fputc(ops['l'] ? '\n' : ops['0'] ? '\0' : ' ',fout);
 	}
@@ -2036,7 +2183,7 @@ int bin_read(name,args,ops,func) /**/
 char *name;char **args;char *ops;int func;
 {
 char *reply,*pmpt;
-int bsiz,c,gotnl = 0;
+int bsiz,c = 0,gotnl = 0;
 char *buf,*bptr;
 
 	reply = (*args) ? *args++ : "REPLY";
@@ -2065,7 +2212,6 @@ char *buf,*bptr;
 		(full(bufstack)) ? (char *) getnode(bufstack) : NULL;
 	while (*args) {
 		buf = bptr = zalloc(bsiz = 64);
-redo:
 		for(;;) {
 			if (gotnl) break;
 			c = zread();
@@ -2073,7 +2219,8 @@ redo:
 				bptr--;
 				continue;
 			}
-			if (c == EOF || isep(c) || c == '\n') break;
+			if (c == EOF || (isep(c) && bptr != buf) || c == '\n') break;
+			if (isep(c)) continue;
 			*bptr++ = c;
 			if (bptr == buf+bsiz) {
 				buf = realloc(buf,bsiz *= 2);
@@ -2089,7 +2236,6 @@ redo:
 			return 1;
 		}
 		if (c == '\n') gotnl = 1;
-		if (bptr == buf) goto redo;
 		*bptr = '\0';
 		setsparam(reply,buf);
 		reply = *args++;
@@ -2103,12 +2249,14 @@ redo:
 				continue;
 			}
 			if (c == EOF || (c == '\n' && !zbuf)) break;
+			if (isep(c) && bptr == buf) continue;
 			*bptr++ = c;
 			if (bptr == buf+bsiz) {
 				buf = realloc(buf,bsiz *= 2);
 				bptr = buf+(bsiz/2);
 			}
 		}
+	while(bptr > buf && isep(bptr[-1])) bptr--;
 	*bptr = '\0';
 	setsparam(reply,buf);
 	if (c == EOF) {
@@ -2176,7 +2324,7 @@ Lknode n;
 			zerrnam("builtin","command name expected",NULL,0);
 			return 1;
 			}
-		for (t0 = 0, b = builtins; b->name; b++,t0++)
+		for (b = builtins; b->name; b++)
 			if (!strcmp(name,b->name))
 				break;
 		if (!b->name)
@@ -2192,10 +2340,9 @@ Lknode n;
 		optstr = NULL;
 	if (optstr)
 		while (arg &&
-				((sense = *arg == '-') || fset(BINF_PLUSOPTS) && *arg == '+') &&
+				((sense = *arg == '-') || (fset(BINF_PLUSOPTS) && *arg == '+')) &&
 				(fset(BINF_PLUSOPTS) || !atoi(arg)))
 			{
-			pp = arg;
 			if (arg[1] == '-')
 				arg++;
 			if (!arg[1])
@@ -2209,7 +2356,7 @@ Lknode n;
 			op = -1;
 			while (*++arg)
 				if (strchr(b->optstr,op = *arg))
-					ops[*arg] = (sense) ? 1 : 2;
+					ops[(int)*arg] = (sense) ? 1 : 2;
 				else
 					break;
 			if (*arg)
@@ -2234,7 +2381,10 @@ Lknode n;
 						}
 					else
 						{
-						ops[c] = ops['o'];
+						if (c == INTERACTIVE)
+							zerr("can't change option: %s",arg,0);
+						else
+							ops[c] = ops['o'];
 						arg = ugetnode(args);
 						}
 					}
@@ -2263,7 +2413,7 @@ Lknode n;
 		auxdata = "-";
 	if (pp = b->defopts)
 		while (*pp)
-			ops[*pp++] = 1;
+			ops[(int)*pp++] = 1;
 	if (arg)
 		{
 		argc = 1;
@@ -2381,10 +2531,10 @@ char **x;
 		listhtable(paramtab,(HFunc) printparam);
 		}
    for (opp = optns; opp->name; opp++)
-      if (ops[opp->id] == 1)
-         opts[opp->id] = OPT_SET;
-      else if (ops[opp->id] == 2)
-         opts[opp->id] = OPT_UNSET;
+      if (ops[(int)opp->id] == 1)
+         opts[(int)opp->id] = OPT_SET;
+      else if (ops[(int)opp->id] == 2)
+         opts[(int)opp->id] = OPT_UNSET;
 	if (!*argv && !ops['-'])
 		return 0;
 	permalloc();
@@ -2438,7 +2588,7 @@ char *str,optbuf[3],*opch = optbuf+1;
 	}
 	if (zoptind > arrlen(args)) return 1;
 	str = args[zoptind-1];
-	if (*str != '+' && *str != '-' || optcind >= strlen(str) ||
+	if ((*str != '+' && *str != '-') || optcind >= strlen(str) ||
 			!strcmp("--",str)) {
 		if (*str == '+' || *str == '-')
 			zoptind++;
@@ -2565,6 +2715,15 @@ long t0;
 #ifdef RLIMIT_RSS
 		case RLIMIT_RSS: printf("resident set size (kbytes) "); t0 /= 1024; break;
 #endif
+#ifdef RLIMIT_MEMLOCK
+		case RLIMIT_MEMLOCK: printf("locked-in-memory size (kb) "); t0 /= 1024; break;
+#endif
+#ifdef RLIMIT_NPROC
+		case RLIMIT_NPROC: printf("processes                  "); break;
+#endif
+#ifdef RLIMIT_OFILE
+		case RLIMIT_OFILE: printf("open files                 "); break;
+#endif
 #ifdef RLIMIT_NOFILE
 		case RLIMIT_NOFILE: printf("file descriptors           "); break;
 #endif
@@ -2599,6 +2758,18 @@ int res,hard;
 	else if (ops['m'])
 		res = RLIMIT_RSS;
 #endif
+#ifdef RLIMIT_MEMLOCK
+	else if (ops['l'])
+		res = RLIMIT_MEMLOCK;
+#endif
+#ifdef RLIMIT_NPROC
+	else if (ops['p'])
+		res = RLIMIT_NPROC;
+#endif
+#ifdef RLIMIT_OFILE
+	else if (ops['o'])
+		res = RLIMIT_OFILE;
+#endif
 #ifdef RLIMIT_NOFILE
 	else if (ops['n'])
 		res = RLIMIT_NOFILE;
@@ -2635,6 +2806,9 @@ int res,hard;
 			case RLIMIT_DATA: case RLIMIT_STACK:
 #ifdef RLIMIT_RSS
 			case RLIMIT_RSS:
+#endif
+#ifdef RLIMIT_MEMLOCK
+			case RLIMIT_MEMLOCK:
 #endif
 				t0 *= 1024; break;
 			}
@@ -2900,11 +3074,11 @@ char *usrkeys = NULL;
 		case 'b': flags |= CC_BINDINGS; break;
 		case 'k':
 			flags |= CC_USRKEYS;
-			if ((*argv)[1]) { usrkeys = (*argv)+1; *argv = ""-1; }
+			if ((*argv)[1]) { usrkeys = (*argv)+1; *argv = "\0"; }
 			else if (!argv[1]) {
 				zerrnam(name,"variable name expected after -k",NULL,0);
 				return 1;
-			} else { usrkeys = *++argv; *argv = ""-1; }
+			} else { usrkeys = *++argv; *argv = "\0"; }
 			break;
 		case 'C': cc = &cc_compos; break;
 		case 'D': cc = &cc_default; break;
@@ -2966,4 +3140,3 @@ char *name;char **argv;char *ops;int func;
 	else if (ops['u']) ttyfrozen = 0;
 	return 0;
 }
-
