@@ -20,14 +20,9 @@
 
 #include "zsh.h"
 #include <sys/errno.h>
-#ifdef __hpux
-#include <ndir.h>
-#else
-#include <sys/dir.h>
-#endif
 
-#define execerr() { if (forked) exit(1); \
-	closemnodes(mfds); errflag = 1; return; }
+#define execerr() { if (forked) _exit(1); \
+	closemnodes(mfds); lastval = 1; return; }
 
 static Lklist args;
 static Cmdnam cn;
@@ -171,20 +166,26 @@ char **eep;
 void execute(dash) /**/
 int dash;
 {
+static Lklist exargs;
 char **argv,*arg0,**pp;
 char *z,*s,buf[MAXCMDLEN],buf2[MAXCMDLEN];
-int ee = 0,eno = 0;
+int ee,eno = 0;
 
-	if (!full(args)) {
+	if (empty(args)) {
 		zerr("no command",NULL,0);
 		_exit(1);
 	}
-	cn = (Cmdnam) gethnode(peekfirst(args),cmdnamtab);
+	if (!exargs && (s = zgetenv("STTY"))) {
+		exargs = args;
+		args = (Lklist) 0;
+		zyztem("stty",s);
+		args = exargs;
+		exargs = (Lklist) 0;
+	}
+	arg0 = peekfirst(args);
+	cn = (Cmdnam) gethnode(arg0,cmdnamtab);
 	if (cn && cn->type == DISABLED)
 		cn = NULL;
-	if (s = zgetenv("STTY"))
-		zyztem("stty",s);
-	arg0 = peekfirst(args);
 	if (z = zgetenv("ARGV0")) {
 		setdata(firstnode(args),ztrdup(z));
 		delenv(z-6);
@@ -365,8 +366,10 @@ List list;
 				dotrap(SIGDEBUG);
 			if (sigtrapped[SIGERR] && lastval)
 				dotrap(SIGERR);
-			if (list->right && !retflag)
+			if (list->right && !retflag) {
+				/* errflag = 0; */
 				execlist(list->right);
+			}
 			break;
 	}
 }
@@ -648,18 +651,20 @@ Lklist vl;
 		postfork(vl,1);
 		if (errflag)
 			return;
-		if (v->type == PMFLAG_s && (!full(vl) || !nextnode(firstnode(vl)))) {
+		if (v->type == PMFLAG_s && (empty(vl) || !nextnode(firstnode(vl)))) {
 			Param pm;
 			char *val;
 
-			if (!full(vl))
-				pm = setsparam(v->name,val = ztrdup(""));
+			if (empty(vl))
+				val = ztrdup("");
 			else {
 				untokenize(peekfirst(vl));
-				pm = setsparam(v->name,val = ztrdup(ugetnode(vl)));
+				val = ztrdup(ugetnode(vl));
 			}
+			pm = setsparam(v->name,ztrdup(val));
 			if (export && !(pm->flags & PMFLAG_x))
 				addenv(v->name,val);
+			free(val);
 			continue;
 		}
 		ptr = arr = (char **) zalloc(sizeof(char **)*(countnodes(vl)+1));
@@ -680,20 +685,20 @@ void execcmd(cmd,input,output,bkg,last1) /**/
 Cmd cmd;int input;int output;int bkg;int last1;
 {
 int type;
-long pid = 0;
+long pid;
 int save[10],t0;
 struct redir *fn;
 struct multio *mfds[10];
-int fil,forked = 0,iscursh = 0,nullexec = 0;
+int fil,forked = 0,iscursh,nullexec = 0;
 char *text;
 
 	args = cmd->args;
 	cn = NULL;
 	for (t0 = 0; t0 != 10; t0++) {
-		save[t0] = 0;
+		save[t0] = -1;
 		mfds[t0] = NULL;
 	}
-	if ((type = cmd->type) == SIMPLE && !full(args))
+	if ((type = cmd->type) == SIMPLE && empty(args))
 		if (full(cmd->redir))
 			if (cmd->flags & CFLAG_EXEC) {
 				nullexec = 1;
@@ -709,36 +714,44 @@ char *text;
 				addnode(args,strdup(nullcmd));
 		else {
 			addvars(cmd->vars,0);
+			if (errflag)
+				lastval = 1;
 			return;
 		}
 	if (full(args) && *(char *) peekfirst(args) == '%') {
 		insnode(args,(Lknode) args,strdup((bkg) ? "bg" : "fg"));
 		bkg = 0;
 	}
-	if (isset(AUTORESUME) && !bkg && !full(cmd->redir) && full(args) &&
+	if (isset(AUTORESUME) && !bkg && empty(cmd->redir) && full(args) &&
 			!input && type == SIMPLE && !nextnode(firstnode(args))) {
 		if (unset(NOTIFY)) scanjobs();
 		if (findjobnam(peekfirst(args)) != -1)
 			pushnode(args,strdup("fg"));
 	}
 	if (unset(RMSTARSILENT) && interact && isset(SHINSTDIN) &&
-				type == SIMPLE && full(args) && nextnode(firstnode(args)) &&
-				!strcmp(peekfirst(args),"rm")) {
-			char *s = getdata(nextnode(firstnode(args)));
+			type == SIMPLE && full(args) && nextnode(firstnode(args)) &&
+			!strcmp(peekfirst(args),"rm") &&
+			!(cmd->flags & CFLAG_NOGLOB)) {
+		Lknode node, next;
+
+		for (node = nextnode(firstnode(args)); node && !errflag; node = next) {
+			char *s = getdata(node);
 			int l = strlen(s);
 
-			if (s[0] == Star && !s[1])
-				checkrmall(pwd);
-			else if (l > 2 && s[l-2] == '/' && s[l-1] == Star) {
+			next = nextnode(node);
+			if (s[0] == Star && !s[1]) {
+				if (!checkrmall(pwd)) uremnode(args,node);
+			} else if (l > 2 && s[l-2] == '/' && s[l-1] == Star) {
 				char t = s[l-2];
 				s[l-2] = 0;
-				checkrmall(s);
+				if (!checkrmall(s)) uremnode(args,node);
 				s[l-2] = t;
 			}
+		}
+		if (!nextnode(firstnode(args))) errflag = 1;
 	}
 	if (jobbing) {	/* get the text associated with this command */
-		char *s;
-		s = text = getjobtext((vptr) cmd);
+		text = getjobtext((vptr) cmd);
 	} else text = NULL;
 	prefork(args);	/* do prefork substitutions */
 	if (errflag) {
@@ -757,7 +770,7 @@ char *text;
 		}
 	}
 	if (type == SIMPLE && !cn && isset(AUTOCD) && isset(SHINSTDIN) &&
-			full(args) && !full(cmd->redir) &&
+			full(args) && empty(cmd->redir) &&
 			!nextnode(firstnode(args)) && cancd(peekfirst(args))) {
 		pushnode(args,strdup("cd"));
 		cn = (Cmdnam) gethnode("cd",cmdnamtab);
@@ -822,22 +835,30 @@ char *text;
 	spawnpipes(cmd->redir);		/* do process substitutions */
 	while (full(cmd->redir))
 		if ((fn = (struct redir*) ugetnode(cmd->redir))->type == INPIPE) {
-			if (fn->fd2 == -1)
+			if (fn->fd2 == -1) {
+				fixfds(save);
 				execerr();
+			}
 			addfd(forked,save,mfds,fn->fd1,fn->fd2,0);
 		} else if (fn->type == OUTPIPE) {
-			if (fn->fd2 == -1)
+			if (fn->fd2 == -1) {
+				fixfds(save);
 				execerr();
+			}
 			addfd(forked,save,mfds,fn->fd1,fn->fd2,1);
 		} else {
 			if (!(fn->type == HERESTR || fn->type == CLOSE || fn->type ==
 					MERGE || fn->type == MERGEOUT))
 				if (xpandredir(fn,cmd->redir))
 					continue;
-			if (errflag) execerr();
+			if (errflag) {
+				fixfds(save);
+				execerr();
+			}
 			if (fn->type == HERESTR) {
 				fil = getherestr(fn);
 				if (fil == -1) {
+					fixfds(save);
 					if (errno != EINTR)
 						zerr("%e",NULL,errno);
 					execerr();
@@ -846,6 +867,7 @@ char *text;
 			} else if (fn->type == READ) {
 				fil = open(fn->name,O_RDONLY);
 				if (fil == -1) {
+					fixfds(save);
 					if (errno != EINTR)
 						zerr("%e: %s",fn->name,errno);
 					execerr();
@@ -859,10 +881,15 @@ char *text;
 			} else if (fn->type == MERGE || fn->type == MERGEOUT) {
 				if (fn->fd2 == FD_COPROC)
 					fn->fd2 = (fn->type == MERGEOUT) ? coprocout : coprocin;
-				if (!forked && fn->fd1 < 10)
-					save[fn->fd1] = movefd(fn->fd1);
 				closemn(mfds,fn->fd1);
 				fil = dup(fn->fd2);
+				if (fil == -1) {
+					char fdstr[4];
+					fixfds(save);
+					sprintf(fdstr,"%d",fn->fd2);
+					zerr("%s: %e",fdstr,errno);
+					execerr();
+				}
 				addfd(forked,save,mfds,fn->fd1,fil,fn->type == MERGEOUT);
 			} else {
 				if (fn->type >= APP)
@@ -873,6 +900,7 @@ char *text;
 					fil = open(fn->name,dontclob(fn) ? 
 						O_WRONLY|O_CREAT|O_EXCL : O_WRONLY|O_CREAT|O_TRUNC,0666);
 				if (fil == -1) {
+					fixfds(save);
 					if (errno != EINTR)
 						zerr("%e: %s",fn->name,errno);
 					execerr();
@@ -886,8 +914,12 @@ char *text;
 	for (t0 = 0; t0 != 10; t0++)
 		closemn(mfds,t0);
 
-	if (nullexec)
+	if (nullexec) {
+		for (t0 = 0; t0 != 10; t0++)
+			if (save[t0] != -1)
+				close(save[t0]);
 		return;
+	}
 	if (unset(NOEXEC))
 		if (type >= CURSH)
 			{
@@ -900,10 +932,19 @@ char *text;
 			}
 		else if (iscursh)		/* builtin or shell function */
 			{
-			if (cmd->vars)
+			if (!cn) {
+				lastval = 1;
+				return;
+			}
+			if (cmd->vars) {
 				addvars(cmd->vars,0);
+				if (errflag) {
+					lastval = 1;
+					return;
+				}
+			}
 			fixcline(args);
-			if (cn && cn->type == SHFUNC)
+			if (cn->type == SHFUNC)
 				execshfunc(cmd,cn);
 			else
 				{
@@ -923,8 +964,13 @@ char *text;
 			}
 		else
 			{
-			if (cmd->vars)
+			if (cmd->vars) {
 				addvars(cmd->vars,1);
+				if (errflag) {
+					lastval = 1;
+					return;
+				}
+			}
 			if (type == SIMPLE)
 				{
 				closem();
@@ -944,11 +990,13 @@ err:
 void fixfds(save) /**/
 int *save;
 {
+int old_errno = errno;
 int t0;
 
 	for (t0 = 0; t0 != 10; t0++)
-		if (save[t0])
+		if (save[t0] != -1)
 			redup(save[t0],t0);
+	errno = old_errno;
 }
 
 void entersubsh(bkg) /**/
@@ -991,10 +1039,10 @@ int bkg;
 	if (interact)
 		{
 		signal(SIGTERM,SIG_DFL);
-		if (sigtrapped[SIGINT])
-			signal(SIGINT,SIG_IGN);
+		if (sigtrapped[SIGINT] != 2)
+			signal(SIGINT, SIG_DFL);
 		}
-	if (!sigtrapped[SIGQUIT])
+	if (sigtrapped[SIGQUIT] != 2)
 		signal(SIGQUIT,SIG_DFL);
 	opts[MONITOR] = OPT_UNSET;
 	clearjobtab();
@@ -1020,8 +1068,11 @@ int qt = 0,siz = 0,l,strip = 0;
 char *s,*t,*bptr;
 
 	for (s = str; *s; s++)
-		if (*s == Nularg)
+		if (INULL(*s))
+			{
+			*s = Nularg;
 			qt = 1;
+			}
 	untokenize(str);
 	if (typ == HEREDOCDASH)
 		{
@@ -1034,7 +1085,7 @@ char *s,*t,*bptr;
 		{
 		char *u,*v;
 
-		if (!hgets(pbuf,256))
+		if (!hgets(pbuf,sizeof(pbuf)))
 			break;
 		bptr = pbuf;
 		if (strip)
@@ -1189,6 +1240,8 @@ char *cmd;int qt;
 {
 List list;
 int pipes[2];
+int pid;
+int status;
 
 	if (*cmd == '<') {
 		int stream;
@@ -1219,12 +1272,38 @@ int pipes[2];
 	if (!(list = parselstring(cmd)))
 		return NULL;
 	mpipe(pipes);
-	if (phork())
+	if ((cmdoutpid = pid = phork()) > 0)
 		{
+		Lklist retval;
+
 		popheap();
 		close(pipes[1]);
-		/* chldsuspend(); */
-		return readoutput(pipes[0],qt);
+#ifdef WAITPID
+		retval =  readoutput(pipes[0],qt);
+		waitpid(pid,&status,0);
+#else
+		blockchld();
+		retval =  readoutput(pipes[0],qt);
+		if (kill(pid, 0) >= 0)
+			{
+			unblockchld();
+			chldsuspend();
+			}
+		else
+			unblockchld();
+#endif
+		lastval = cmdoutval;
+		cmdoutval = 0;
+		return retval;
+		}
+ 	else if (pid == -1)
+		{
+		popheap();
+		close(pipes[0]);
+		close(pipes[1]);
+		errflag = 1;
+		cmdoutpid = 0;
+		return NULL;
 		}
 	subsh = 1;
 	close(pipes[0]);
@@ -1234,7 +1313,9 @@ int pipes[2];
 	exiting = 1;
 	execlist(list);
 	close(1);
-	exit(0);  return NULL;
+	_exit(lastval);
+	zerr("exit returned in child!!",NULL,0);
+	kill(getpid(),SIGKILL);
 }
 
 /* read output of command substitution */
@@ -1250,6 +1331,10 @@ FILE *fin;
 	fin = fdopen(in,"r");
 	ret = newlist();
 	ptr = buf = ncalloc(bsiz = 64);
+	if (qt) {
+		*ptr++ = Nularg;
+		cnt++;
+	}
 	while ((c = fgetc(fin)) != EOF)
 		if (!qt && isep(c)) {
 			if (cnt) {
@@ -1257,7 +1342,6 @@ FILE *fin;
 				addnode(ret,buf);
 				ptr = buf = ncalloc(bsiz = 64);
 				cnt = 0;
-				ptr = buf;
 			}
 		} else {
 			*ptr++ = c;
@@ -1286,7 +1370,6 @@ char *cmd;
 int pid;
 #endif
 char *nam = gettemp(),*str;
-int tfil;
 List list;
 
 	if (thisjob == -1)
@@ -1318,11 +1401,13 @@ List list;
 	subsh = 1;
 	close(1);
 	entersubsh(0);
-	tfil = creat(nam,0666);
+	(void) creat(nam,0666);
 	exiting = 1;
 	execlist(list);
 	close(1);
-	exit(0); return NULL;
+	_exit(lastval);
+	zerr("exit returned in child!!",NULL,0);
+	kill(getpid(),SIGKILL);
 }
 
 /* get a temporary named pipe */
@@ -1378,7 +1463,7 @@ char *pnam,*str;
 	if (fd == -1)
 		{
 		zerr("can't open %s: %e",pnam,errno);
-		_exit(0);
+		_exit(1);
 		}
 	redup(fd,1);
 	fd = open("/dev/null",O_RDONLY);
@@ -1386,7 +1471,7 @@ char *pnam,*str;
 	exiting = 1;
 	execlist(list);
 	close(1);
-	_exit(0);  return NULL;
+	_exit(lastval);  return NULL;
 #endif
 }
 
@@ -1429,7 +1514,7 @@ char *pnam,*str;
 	redup(fd,0);
 	exiting = 1;
 	execlist(list);
-	_exit(0);  return NULL;
+	_exit(lastval);  return NULL;
 #endif
 }
 
@@ -1440,7 +1525,7 @@ char *cmd;
 {
 List list;
 int pipes[2];
-char *str = cmd;
+char *str;
 
 	for (str = cmd; *str && *str != Outpar; str++);
 	if (!*str)
@@ -1461,7 +1546,7 @@ char *str = cmd;
 	redup(pipes[1],1);
 	exiting = 1;
 	execlist(list);
-	_exit(0);  return 0;
+	_exit(lastval);  return 0;
 }
 
 /* < <(...) */
@@ -1488,12 +1573,12 @@ char *str;
 		return pipes[1];
 		}
 	close(pipes[1]);
-	closem();
 	entersubsh(1);
 	redup(pipes[0],0);
+	closem();
 	exiting = 1;
 	execlist(list);
-	_exit(0);  return 0;
+	_exit(lastval);  return 0;
 }
 
 /* run a list, saving the current job num */
@@ -1517,16 +1602,29 @@ char *gettemp() /**/
 char *zgetwd() /**/
 {
 static char buf0[MAXPATHLEN];
-char buf3[MAXPATHLEN],*buf2 = buf0+1;
+char buf3[MAXPATHLEN];
+#ifdef apollo
+char *buf2 = buf0+2; /* changed +1 to +2  RBC 17.11.91 */
+#else
+char *buf2 = buf0+1;
+#endif
 struct stat sbuf;
 struct direct *de;
 DIR *dir;
-ino_t ino = -1;
-dev_t dev = -1;
+ino_t ino, rootino = ~0;
+dev_t dev, rootdev = ~0;
 
 	holdintr();
 	buf2[0] = '\0';
 	buf0[0] = '/';
+#ifdef apollo	
+	buf0[1] = '/'; /* added RBC 17.11.91 */
+#endif	
+	if (stat(buf0,&sbuf) >= 0)
+		{
+		rootino = sbuf.st_ino;
+		rootdev = sbuf.st_dev;
+		}
 	for(;;)
 		{
 		if (stat(".",&sbuf) < 0)
@@ -1543,7 +1641,8 @@ dev_t dev = -1;
 			noholdintr();
 			return ztrdup(".");
 			}
-		if (sbuf.st_ino == ino && sbuf.st_dev == dev)
+		if ((sbuf.st_ino == ino && sbuf.st_dev == dev) ||
+		    (ino == rootino && dev == rootdev))
 			{
 			chdir(buf0);
 			noholdintr();
@@ -1657,7 +1756,10 @@ char *s;
 			int t0 = getsignum(s+4);
 
 			if (t0 != -1)
+				{
 				settrap(t0,cmd->u.list);
+				permalloc();
+				}
 			}
 		}
 	heapalloc();
@@ -1738,7 +1840,7 @@ List xexitfn;
 	while (s = getnode(locallist)) unsetparam(s);
 	free(locallist);
 	locallist = olist;
-	retflag = 0;
+	breaks = retflag = 0;
 	freearray(pparams);
 	free(argzero);
 	argzero = oargv0;
@@ -1794,12 +1896,12 @@ int fd;
 
 /* check to see if AUTOCD applies here */
 
-int cancd(s)
+int cancd(s) /**/
 char *s;
 {
 char *t;
 
-	if ((t = getsparam(s)) && *t == '/') return 1;
+	if (isset(CDABLEVARS) && (t = getsparam(s)) && *t == '/') return cancd2(t);
 	if (*s != '/')
 		{
 		char sbuf[MAXPATHLEN],**cp;
@@ -1819,11 +1921,10 @@ char *t;
 	return cancd2(s);
 }
 
-int cancd2(s)
+int cancd2(s) /**/
 char *s;
 {
 struct stat buf;
 
 	return !(access(s,X_OK) || stat(s,&buf) || !S_ISDIR(buf.st_mode));
 }
-
