@@ -1,5 +1,7 @@
 struct asgment;
 struct utmp;
+struct hp;
+typedef struct hp *Hp;
 
 #include "builtin.pro"
 #include "cond.pro"
