@@ -20,15 +20,6 @@
 
 #include "zsh.h"
 
-#ifdef __hpux
-#include <ndir.h>
-#else
-#ifdef SYSV
-#define direct dirent
-#else
-#include <sys/dir.h>
-#endif
-#endif
 #include <sys/errno.h>
 
 #define exists(X) (access(X,0) == 0 || readlink(X,NULL,0) == 0)
@@ -41,6 +32,9 @@ static char pathbuf[MAXPATHLEN];	/* pathname buffer */
 static char **matchbuf;		/* array of matches */
 static char **matchptr;		/* &matchbuf[matchct] */
 static Comp exclude;			/* pattern to exclude */
+#ifdef ULTRIX
+typedef struct stat *Statptr;   /* This makes the Ultrix compiler happy.  Go figure. */
+#endif
 
 /* max # of qualifiers */
 
@@ -49,7 +43,10 @@ static Comp exclude;			/* pattern to exclude */
 static int (*qualfuncs[QUALCT])DCLPROTO((struct stat *,long));
 static long qualdata[QUALCT];
 static int qualsense[QUALCT];
+static int qualamc[QUALCT];
+static int qualrange[QUALCT];
 static int qualct;
+static int range, amc;
 static int gf_nullglob,gf_markdirs,gf_noglobdots;
 
 /* pathname component in filename patterns */
@@ -70,13 +67,15 @@ Lklist list;Lknode *np;
 {
 Lknode node = prevnode(*np);
 Lknode next = nextnode(*np);
+char *str;		/* the pattern */
 int sl;			/* length of the pattern */
-char *ostr;		/* the pattern before the parser chops it up */
 Complist q;		/* pattern after parsing */
-char *str = getdata(*np);	/* the pattern */
+char *ostr = getdata(*np);		/* the pattern before the parser chops it up */
 
+	heapalloc();
+	str = strdup(ostr);
+	lastalloc();
 	sl = strlen(str);
-	ostr = strdup(str);
 	uremnode(list,*np);
 	qualct = 0;
 	gf_nullglob = isset(NULLGLOB);
@@ -86,8 +85,12 @@ char *str = getdata(*np);	/* the pattern */
 		{
 		char *s;
 		int sense = 0;
-		long data;
+		long data = 0;
+#ifdef ULTRIX
+		int (*func) DCLPROTO((Statptr,long));
+#else
 		int (*func) DCLPROTO((struct stat *,long));
+#endif
 
 		for (s = str+sl-2; s != str; s--)
 			if (*s == Bar || *s == Outpar || *s == Inpar)
@@ -95,10 +98,13 @@ char *str = getdata(*np);	/* the pattern */
 		if (*s == Inpar)
 			{
 			*s++ = '\0';
-			func = NULL;
 			while (*s != Outpar)
 				{
-				func = NULL;
+#ifdef ULTRIX
+				func = (int (*) DCLPROTO((Statptr,long))) 0;
+#else
+				func = (int (*) DCLPROTO((struct stat *,long))) 0;
+#endif
 				if (idigit(*s))
 					{
 					func = qualflags;
@@ -108,7 +114,7 @@ char *str = getdata(*np);	/* the pattern */
 					}
 				else switch ((int)(unsigned char)(*s++))
 					{
-					case (int)(unsigned char)Hat: case '^': sense = 1-sense; break;
+					case (int)STOUC(Hat): case '^': sense = 1-sense; break;
 #ifdef S_IFLNK
 					case '@': func = qualmode; data = S_IFLNK; break;
 #endif
@@ -121,7 +127,7 @@ char *str = getdata(*np);	/* the pattern */
 					case '/': func = qualmode; data = S_IFDIR; break;
 					case '.': func = qualmode; data = S_IFREG; break;
 					case '%': func = qualisdev; break;
-					case (int)(unsigned char)Star:  func = qualiscom; break;
+					case (int)STOUC(Star):  func = qualiscom; break;
 					case 'R': func = qualflags; data = 0004; break;
 					case 'W': func = qualflags; data = 0002; break;
 					case 'X': func = qualflags; data = 0001; break;
@@ -139,6 +145,14 @@ char *str = getdata(*np);	/* the pattern */
 					case 'M': gf_markdirs = !sense; break;
 					case 'N': gf_nullglob = !sense; break;
 					case 'D': gf_noglobdots = sense; break;
+					case 'a': amc = 0; func = qualtime; goto getrange;
+					case 'm': amc = 1; func = qualtime; goto getrange;
+					case 'c': amc = 2; func = qualtime; goto getrange;
+					case 'L': func = qualsize; 
+getrange:
+							if (range = *s == '+' ? 1 : *s == '-' ? -1 : 0) ++s;
+							data = qgetnum(&s);
+							break;
 					default: zerr("unknown file attribute",NULL,0); return;
 					}
 				if (func)
@@ -151,6 +165,8 @@ char *str = getdata(*np);	/* the pattern */
 					qualfuncs[qualct] = func;
 					qualsense[qualct] = sense;
 					qualdata[qualct] = data;
+					qualrange[qualct] = range;
+					qualamc[qualct] = amc;
 					qualct++;
 					}
 				if (errflag)
@@ -168,7 +184,11 @@ char *str = getdata(*np);	/* the pattern */
 		qualsense[0] = 0;
 		qualct = 1;
 		}
-	qualfuncs[qualct] = NULL;
+#ifdef ULTRIX
+	qualfuncs[qualct] = (int (*) DCLPROTO((Statptr,long))) 0;
+#else
+	qualfuncs[qualct] = (int (*) DCLPROTO((struct stat *,long))) 0;
+#endif
 	if (*str == '/')	/* pattern has absolute path */
 		{
 		str++;
@@ -205,7 +225,8 @@ char *str = getdata(*np);	/* the pattern */
 			*matchptr++ = strdup(ostr);
 			matchct = 1;
 		}
-	qsort(&matchbuf[0],matchct,sizeof(char *),notstrcmp);
+	qsort((vptr)&matchbuf[0],matchct,sizeof(char *),
+			(int (*) DCLPROTO((const void *, const void *)))notstrcmp);
 	matchptr = matchbuf;
 	while (matchct--)			/* insert matches in the arg list */
 		insnode(list,node,*matchptr++);
@@ -234,13 +255,18 @@ int notstrcmp(a,b) /**/
 char **a;char **b;
 {
 char *c = *b,*d = *a;
-int x1,x2;
+int x1,x2,cmp;
 
 	for (; *c == *d && *c; c++,d++);
-	x1 = atoi(c); x2 = atoi(d);
-	if (x1==x2 || unset(NUMERICGLOBSORT))
-		return ((int) (unsigned char) *c-(int) (unsigned char) *d);
-	return x1-x2;
+	cmp = (int) (unsigned char) *c-(int) (unsigned char) *d;
+	if (isset(NUMERICGLOBSORT)) {
+		for (; c > *b && idigit(c[-1]); c--,d--);
+		if(idigit(*c) && idigit(*d)) {
+			x1 = atoi(c); x2 = atoi(d);
+			if(x1 != x2) return x1-x2;
+		}
+	}
+	return cmp;
 }
 
 int forstrcmp(a,b) /**/
@@ -273,13 +299,23 @@ int statted = 0;
 		statted = 1;
 	}
 	if (qualct)	{ /* do the (X) (^X) stuff */
+#ifdef ULTRIX
+		int (**fptr)DCLPROTO((Statptr,long)) = qualfuncs;
+#else
 		int (**fptr)DCLPROTO((struct stat *,long)) = qualfuncs;
+#endif
 		int *sptr = qualsense;
+		int *rptr = qualrange;
+		int *aptr = qualamc;
 		long *lptr = qualdata;
-		struct stat buf;
 
 		if (statted || lstat(s,&buf) >= 0)
-			while (*fptr) if (!(!!((*fptr++)(&buf,*lptr++)) ^ *sptr++)) return;
+			while (*fptr) {
+				range = *rptr++;
+				amc = *aptr++;
+				if (!(!!((*fptr++)(&buf,*lptr++)) ^ *sptr++))
+					return;
+			}
 	}
 	*matchptr++ = s;
 	if (++matchct == matchsz) {
@@ -449,9 +485,9 @@ int prev, bc, comma;
 		memset(ccl, 0, sizeof(ccl) / sizeof(ccl[0]));
 		for (p = str + 1, lastch = 0; p < str2; ) {
 			if (itok(c1 = *p++))
-				c1 = ztokens[c1 - (unsigned char)Pound];
+				c1 = ztokens[c1 - STOUC(Pound)];
 			if (itok(c2 = *p))
-				c2 = ztokens[c2 - (unsigned char)Pound];
+				c2 = ztokens[c2 - STOUC(Pound)];
 			if (c1 == '-' && lastch && p < str2 && lastch <= c2) {
 				while (lastch < c2)
 					ccl[lastch++] = 1;
@@ -513,8 +549,8 @@ int prev, bc, comma;
 		char *zz,*str4;
 		int cnt;
 		
-		for (str4 = str, cnt = 0; cnt || *str != Comma && *str !=
-				Outbrace; str++)
+		for (str4 = str, cnt = 0; cnt || (*str != Comma && *str !=
+				Outbrace); str++)
 			if (*str == Inbrace)
 				cnt++;
 			else if (*str == Outbrace)
@@ -1209,7 +1245,7 @@ int nl = *s;
 char *t = s;
 
 	while (*s)
-		if (*s == Nularg)
+		if (INULL(*s))
 			chuck(s);
 		else
 			s++;
@@ -1271,3 +1307,24 @@ struct stat *buf;long mod;
 	return (buf->st_mode & (S_IFMT|S_IEXEC)) == (S_IFREG|S_IEXEC);
 }
 
+int qualsize(buf,size) /**/
+struct stat *buf;long size;
+{
+    return(range < 0 ? buf->st_size < size :
+           range > 0 ? buf->st_size > size :
+                       buf->st_size == size);
+}
+
+int qualtime(buf,days) /**/
+struct stat *buf;long days;
+{
+    time_t    now, diff;
+
+    time(&now);
+    diff = now - (amc == 0 ? buf->st_atime : amc == 1 ? buf->st_mtime :
+								buf->st_ctime);
+    diff /= 86400l;
+    return(range < 0 ? diff < days :
+           range > 0 ? diff > days :
+                       diff == days);
+}
