@@ -38,17 +38,17 @@ static int lastc;
 void hwaddc(c) /**/
 int c;
 {
-	if (hlastw && hline && (!(errflag || lexstop) || c == HISTSPACE)) {
+	if (hlastw && chline && (!(errflag || lexstop) || c == HISTSPACE)) {
 		if (c == '!' && unset(NOBANGHIST)) hwaddc('\\');
 		*hptr++ = c;
-		if (hptr-hline >= hlinesz) {
+		if (hptr-chline >= hlinesz) {
 			int ll,flag = 0,oldsiz = hlinesz;
 
 			ll = hptr-hlastw;
-			if (curhistent->lex == hline) flag = 1;
-			hline = hp_realloc(&hp_lex,hline,oldsiz,hlinesz = oldsiz+16);
-			if (flag) curhistent->lex = hline;
-			hptr = hline+oldsiz;
+			if (curhistent->lex == chline) flag = 1;
+			chline = hp_realloc(&hp_lex,chline,oldsiz,hlinesz = oldsiz+16);
+			if (flag) curhistent->lex = chline;
+			hptr = chline+oldsiz;
 			hlastw = hptr-ll;
 		}
 	}
@@ -76,7 +76,6 @@ tailrec:
 		isfirstch = 0;
 		hungetch(hatchar);
 		hungets(":s");
-		c = bangchar;
 		goto hatskip;
 		}
 	if (c != ' ')
@@ -108,7 +107,7 @@ hatskip:
 		stophist = 1;
 		goto tailrec;
 		}
-	if (!cflag && inblank(c) || c == '=' || c == '(' || lexstop)
+	if ((!cflag && inblank(c)) || c == '=' || c == '(' || lexstop)
 		{
 		if (lexstop)
 			lexstop = 0;
@@ -169,7 +168,7 @@ hatskip:
 			ev = defev;
 		else if (t0 = atoi(buf))
 			ev = (t0 < 0) ? curhist+t0 : t0;
-		else if (*buf == bangchar)
+		else if ((unsigned) *buf == bangchar)
 			ev = curhist-1;
 		else if (*buf == '#')
 			ev = curhist;
@@ -331,8 +330,7 @@ hatskip:
 				default:
 					herrflush();
 					zerr("illegal modifier: %c",NULL,c);
-				habort();
-				break;
+					habort();
 			}
 		}
 	else
@@ -382,7 +380,7 @@ Alias ix;
 
 int hgetch() /**/
 {
-unsigned char *line,*pmpt,*pmpt2 = NULL;
+unsigned char *line,*pmpt = NULL,*pmpt2 = NULL;
 int plen;
 
 start:
@@ -423,13 +421,13 @@ start:
 		}
 	if (interact && isset(SHINSTDIN))
 		if (!isfirstln)
-			pmpt = (unsigned char *)putprompt(prompt2,&plen);
+			pmpt = (unsigned char *)putprompt(prompt2,&plen,0);
 		else
 			{
 			int foo;
 
-			pmpt = (unsigned char *)putprompt(prompt,&plen);
-			pmpt2 = (unsigned char *)((rprompt) ? putprompt(rprompt,&foo) : NULL);
+			pmpt = (unsigned char *)putprompt(prompt,&plen,0);
+			pmpt2 = (unsigned char *)((rprompt) ? putprompt(rprompt,&foo,0) : NULL);
 			}
 	if (!(interact && isset(SHINSTDIN) && SHTTY != -1 && isset(USEZLE))) {
 		char *lbuf;
@@ -530,7 +528,7 @@ int slen = strlen(str);
 	inbufct += slen;
 }
 
-/* unget a char and remove it from hline */
+/* unget a char and remove it from chline */
 
 void hungetc(c) /**/
 int c;
@@ -630,9 +628,9 @@ void hbegin() /**/
 {
 	isfirstln = isfirstch = 1;
 	histremmed = errflag = histdone = spaceflag = 0;
-	stophist = isset(NOBANGHIST) || unset(SHINSTDIN);
+	stophist = !interact || isset(NOBANGHIST) || unset(SHINSTDIN);
 	lithist = isset(HISTLIT);
-	hline = hptr = hp_alloc(&hp_lex,hlinesz = 16);
+	chline = hptr = hp_alloc(&hp_lex,hlinesz = 16);
 	curhistent = gethistent(curhist);
 	if (!curhistent->ftim) curhistent->ftim = time(NULL);
 	if (interact && isset(SHINSTDIN) && !strin) {
@@ -643,7 +641,7 @@ void hbegin() /**/
 		curhistent = gethistent(curhist);
 		hp_purge(hp_lex,curhist-histsiz);
 		hp_purge(hp_lit,curhist-lithistsiz);
-		curhistent->lex = hline;
+		curhistent->lex = chline;
 		*(curhistent->lit = hp_alloc(&hp_lit,1)) = '\0';
 	} else
 		histremmed = 1;
@@ -661,15 +659,15 @@ int hend() /**/
 int flag,save = 1;
 Histent he;
 
-	if (!hline)
+	if (!chline)
 		return 1;
 	if (!interact || strin || unset(SHINSTDIN)) {
-		hp_free(hp_lex,hline,hlinesz);
+		hp_free(hp_lex,chline,hlinesz);
 		return 1;
 	}
 	flag = histdone;
 	histdone = 0;
-	if (hptr < hline+2)
+	if (hptr < chline+2)
 		save = 0;
 	else {
 		char *s,*t;
@@ -678,18 +676,18 @@ Histent he;
 		if (*s && *(t = s+strlen(s)-1) == HISTSPACE) *t = '\0';
 		hptr[-1] = '\0';
 		if (hptr[-2] == '\n')
-			if (hline[1]) {
+			if (chline[1]) {
 				if (hptr[-3] == HISTSPACE) hptr[-3] = '\0';
 			} else save = 0;
 		he = gethistent(curhist-1);
-		if (!strcmp(hline,"\n") ||
-				(isset(HISTIGNOREDUPS) && he->lex && !strcmp(he->lex,hline)) ||
+		if (!strcmp(chline,"\n") ||
+				(isset(HISTIGNOREDUPS) && he->lex && !strcmp(he->lex,chline)) ||
 				(isset(HISTIGNORESPACE) && spaceflag))
 			save = 0;
 	}
 	if (flag & (HISTFLAG_DONE|HISTFLAG_RECALL)) {
 		char *ptr,*p;
-		p = ptr = ztrdup(hline);
+		p = ptr = ztrdup(chline);
 		for (;*p;p++) if (*p == HISTSPACE) *p = ' ';
 		if ((flag & (HISTFLAG_DONE|HISTFLAG_RECALL)) == HISTFLAG_DONE) {
 			fprintf(stderr,"%s\n",ptr);
@@ -704,9 +702,10 @@ Histent he;
 	}
 	curhistent->stim = time(NULL);
 	curhistent->ftim = 0L;
+	curhistent->flags = 0;
 	if (!save) remhist();
-	if (hline && !curhistent->lex) hp_free(hp_lex,hline,hlinesz);
-	hline = NULL;
+	if (chline && !curhistent->lex) hp_free(hp_lex,chline,hlinesz);
+	chline = NULL;
 	return !(flag & HISTFLAG_NOEXEC || errflag);
 }
 
@@ -730,7 +729,7 @@ char *hwadd() /**/
 {
 char *ret = hlastw;
 
-	if (hlastw && hline)
+	if (hlastw && chline)
 		{
 		hwaddc(HISTSPACE);
 		if (alstackind || strin)
@@ -1147,13 +1146,13 @@ char *buf = zalloc(bsiz),*ptr;
 
 void inithist() /**/
 {
-	hp_lit = zalloc(sizeof *hp_lit);
+	hp_lit = zcalloc(sizeof *hp_lit);
 	hp_lit->next = NULL;
-	hp_lit->ptr = hp_lit->pool = zalloc(HEAPSIZE);
+	hp_lit->ptr = hp_lit->pool = zcalloc(HEAPSIZE);
 	hp_lit->free = HEAPSIZE;
-	hp_lex = zalloc(sizeof *hp_lex);
+	hp_lex = zcalloc(sizeof *hp_lex);
 	hp_lex->next = NULL;
-	hp_lex->ptr = hp_lex->pool = zalloc(HEAPSIZE);
+	hp_lex->ptr = hp_lex->pool = zcalloc(HEAPSIZE);
 	hp_lex->free = HEAPSIZE;
 	histentct = (lithistsiz > histsiz) ? lithistsiz : histsiz;
 	histentarr = zcalloc(histentct*sizeof *histentarr);
@@ -1254,7 +1253,7 @@ Hp hlast;
 
 	if (!h->next) return;
 	while (h->next) { hlast = h; h = h->next; }
-	if (h->histno <= lim || h->histno == 0) {
+	if (hlast->histno <= lim) {
 #ifdef MEMDEBUG
 		fprintf(stderr,"purging %d\n",lim);
 #endif
@@ -1274,7 +1273,7 @@ time_t tim = time(NULL);
 
 	if (!s) return;
 	if (in = fopen(s,"r")) {
-		while (fgets(buf,1024,in)) {
+		while (fgets(buf,sizeof(buf),in)) {
 			int l = strlen(buf);
 			char *pt = buf;
 
@@ -1282,18 +1281,38 @@ time_t tim = time(NULL);
 				buf[l-1] = '\0';
 				if (l > 1 && buf[l-2] == '\\') {
 					buf[l-2] = '\n';
-					fgets(buf+l-1,1024-(l-1),in);
+					fgets(buf+l-1,sizeof(buf)-(l-1),in);
 					l = strlen(buf);
 				} else break;
 			}
 			for (;*pt;pt++) if (*pt == ' ') *pt = HISTSPACE;
 
 			ent = gethistent(++curhist);
-			ent->lex = hp_alloc(&hp_lex,strlen(buf)+1);
-			strcpy(ent->lex,buf);
-			ent->lit = hp_alloc(&hp_lit,strlen(buf)+1);
-			strcpy(ent->lit,buf);
+			pt = buf;
+			if (*pt == ':') {
+				pt++;
+				ent->stim = atol(pt);
+				for(;*pt != ':' && *pt; pt++)
+					;
+				if (*pt) {
+					pt++;
+					ent->ftim = atol(pt);
+					for(;*pt != ':' && *pt; pt++)
+						;
+					if (*pt) pt++;
+				} else {
+					ent->ftim = tim;
+				}
+				if (ent->stim == 0) ent->stim = tim;
+				if (ent->ftim == 0) ent->ftim = tim;
+			} else {
 			ent->ftim = ent->stim = tim;
+		}
+			ent->lex = hp_alloc(&hp_lex,strlen(pt)+1);
+			strcpy(ent->lex,pt);
+			ent->lit = hp_alloc(&hp_lit,strlen(pt)+1);
+			strcpy(ent->lit,pt);
+			ent->flags = HIST_OLD;
 		}
 		fclose(in);
 	} else if (err)
@@ -1306,14 +1325,24 @@ char *s;int err;int app;
 char *t;
 FILE *out;
 int ev,flag;
+Histent ent;
 
-	if (!s || !interact) return;
+	if (!s || !interact || savehist == 0) return;
 	ev = curhist-savehist+1;
-	flag = (app) ? O_APPEND : O_TRUNC;
+	flag = (app & 1) ? O_APPEND : O_TRUNC;
 	if (ev < firsthist()) ev = firsthist();
 	if (out = fdopen(open(s,O_CREAT|O_WRONLY|flag,0600),"w")) {
 		for (; ev <= curhist; ev++) {
-			t = quietgetevent(ev);
+			ent = gethistent(ev);
+			if (app & 2) {
+				if (ent->flags & HIST_OLD) continue;
+				ent->flags |= HIST_OLD;
+			}
+			t = ((lithist) ? ent->lit : ent->lex);
+			if (isset(EXTENDEDHISTORY)) {
+			  fprintf(out, ":%d:%d:", ent->stim,
+				  ent->ftim);
+			}
 			for (; *t; t++)
 				if (*t == HISTSPACE) fputc(' ',out);
 				else {
@@ -1340,4 +1369,3 @@ Histent ent;
 	} while (ev < curhist);
 	return ev;
 }
-
