@@ -81,6 +81,7 @@ void heapalloc() /**/
 
 static vptr (*lastcalloc) DCLPROTO((int));
 static vptr (*lastncalloc) DCLPROTO((int));
+static int lastuseheap;
 
 /* set default allocation to malloc() */
 
@@ -88,6 +89,7 @@ void permalloc() /**/
 {
 	lastcalloc = alloc;
 	lastncalloc = ncalloc;
+	lastuseheap = useheap;
 	alloc = zcalloc;
 	ncalloc = zalloc;
 	useheap = 0;
@@ -99,6 +101,7 @@ void lastalloc() /**/
 {
 	alloc = lastcalloc;
 	ncalloc = lastncalloc;
+	useheap = lastuseheap;
 }
 
 struct heap {
@@ -228,7 +231,7 @@ vptr ptr;
 }
 
 char *strdup(s) /**/
-char *s;
+const char *s;
 {
 char *t;
 
@@ -240,7 +243,7 @@ char *t;
 }
 
 char *ztrdup(s) /**/
-char *s;
+const char *s;
 {
 char *t;
 
