@@ -19,6 +19,7 @@
  */
 
 #include "zsh.h"
+#include "version.h"
 #include <pwd.h>
 
 #define new(X) (X=(vptr)alloc(sizeof(*(X))))
@@ -27,7 +28,7 @@ static Param argvparam;
 
 struct iparam {
 	struct hashnode *next; int canfree; char *nam; /* hash data */
-	void *value;
+	vptr value;
 	int (*func1)(); /* set func */
 	int (*func2)(); /* get func */
 	int ct;				/* output base or field width */
@@ -66,10 +67,10 @@ static struct iparam pinit[] = {
 	IPDEF2("IFS",ifsgetfn,ifssetfn,0),
 	IPDEF2("_",underscoregetfn,IFN(nullsetfn),PMFLAG_r),
 
-#define IPDEF3(A,B) {NULL,0,A,NULL,IFN(nullsetfn),IFN(strconstgetfn),0,PMFLAG_r|\
-		PMFLAG_SPECIAL,(vptr)B,NULL,NULL}
-	IPDEF3("HOSTTYPE",HOSTTYPE),
-	IPDEF3("VERSION",VERSIONSTR),
+#define IPDEF3(A) {NULL,0,A,NULL,IFN(nullsetfn),IFN(strconstgetfn),0,PMFLAG_r|\
+		PMFLAG_SPECIAL,NULL,NULL,NULL}
+	IPDEF3("HOSTTYPE"),
+	IPDEF3("VERSION"),
 
 #define IPDEF4(A,B) {NULL,0,A,NULL,IFN(nullsetfn),IFN(intvargetfn),10,\
 		PMFLAG_r|PMFLAG_i|PMFLAG_SPECIAL,(vptr)B,NULL,NULL}
@@ -139,21 +140,23 @@ static struct iparam pinit[] = {
 	IPDEF8("MANPATH","manpath",&manpath),
 	IPDEF8("WATCH","watch",&watch),
 	IPDEF8("HOSTS","hosts",&hosts),
+	IPDEF8("PSVAR","psvar",&psvar),
 	IPDEF8("PATH",NULL,NULL),
 
-#define IPDEF9(A,B,C) {NULL,0,A,NULL,IFN(arrvarsetfn),IFN(arrvargetfn),0,\
-		PMFLAG_A|PMFLAG_SPECIAL|C,(vptr)B,NULL,NULL}
-	IPDEF9("cdpath",&cdpath,0),
-	IPDEF9("fignore",&fignore,0),
-	IPDEF9("fpath",&fpath,0),
-	IPDEF9("mailpath",&mailpath,0),
-	IPDEF9("manpath",&manpath,0),
-	IPDEF9("watch",&watch,0),
-	IPDEF9("hosts",&hosts,0),
-	IPDEF9("signals",&sigptr,PMFLAG_r),
-	IPDEF9("argv",&pparams,0),
-	IPDEF9("*",&pparams,0),
-	IPDEF9("@",&pparams,0),
+#define IPDEF9(A,B,C,D) {NULL,0,A,NULL,IFN(arrvarsetfn),IFN(arrvargetfn),0,\
+		PMFLAG_A|PMFLAG_SPECIAL|C,(vptr)B,NULL,D}
+	IPDEF9("cdpath",&cdpath,0,"CDPATH"),
+	IPDEF9("fignore",&fignore,0,"FIGNORE"),
+	IPDEF9("fpath",&fpath,0,"FPATH"),
+	IPDEF9("mailpath",&mailpath,0,"MAILPATH"),
+	IPDEF9("manpath",&manpath,0,"MANPATH"),
+	IPDEF9("watch",&watch,0,"WATCH"),
+	IPDEF9("hosts",&hosts,0,"HOSTS"),
+	IPDEF9("psvar",&psvar,0,"PSVAR"),
+	IPDEF9("signals",&sigptr,PMFLAG_r,NULL),
+	IPDEF9("argv",&pparams,0,NULL),
+	IPDEF9("*",&pparams,0,NULL),
+	IPDEF9("@",&pparams,0,NULL),
 
 #define IPDEF10(A,C,D) {NULL,0,A,NULL,IFN(D),IFN(C),0,\
 		PMFLAG_A|PMFLAG_SPECIAL,NULL,NULL,NULL}
@@ -166,8 +169,11 @@ static struct iparam pinit[] = {
 	};
 struct iparam *ip;
 
-	for (ip = pinit; ip->nam; ip++) addhperm(ip->nam,ip,paramtab,NULL);
+	for (ip = pinit; ip->nam; ip++) addhperm(ip->nam,ip,paramtab,(FFunc) 0);
 	argvparam = gethnode("argv",paramtab);
+
+	((struct iparam *)gethnode("HOSTTYPE", paramtab))->data = ztrdup(HOSTTYPE);
+	((struct iparam *)gethnode("VERSION", paramtab))->data = ztrdup(VERSIONSTR);
 }
 
 static int unsetflag;
@@ -209,19 +215,13 @@ char *s;
 char *ss;
 
 	for (ss = s; *ss; ss++) if (!iident(*ss)) break;
-	if (!*ss || *ss == '[') return 1;
-	if (*s == Quest)
-		*s = '?';
-	else if (*s == Pound)
-		*s = '#';
-	else if (*s == String || *s == Qstring)
-		*s = '$';
-	else if (*s == Star)
-		*s = '*';
-	if (*s == '#' || *s == '-' || *s == '?' || *s == '$' || *s == '_' ||
-		 *s == '!' || *s == '@' || *s == '*')
-		return 1;
-	return 0;
+	if (!*ss) return 1;
+	if (*ss != '[') return 0;
+	(void) mathevalarg(++ss, &ss);
+	if(*ss == ',' || *ss == Comma)
+		(void) mathevalarg(++ss,&ss);
+	if(*ss != ']' || ss[1]) return 0;
+	return 1;
 }
 
 Value getvalue(pptr,bracks) /**/
@@ -368,21 +368,23 @@ char **ss;
 	}
 	ss = v->pm->gets.afn(v->pm);
 	if (v->a < 0) v->a += arrlen(ss);
-	if (v->a < 0 || v->a > arrlen(ss)) return 0;
+	if (v->a < 0 || v->a >= arrlen(ss)) return 0;
 	return atol(ss[v->a]);
 }
 
 void setstrvalue(v,val) /**/
 Value v;char *val;
 {
-char *s;
-
 	if (v->pm->flags & PMFLAG_r) {
 		free(val);
 		return;
 	}
-	if ((s = v->pm->env) && val)
+	if (v->pm->env)
 		v->pm->env = replenv(v->pm->env,val);
+	else if (isset(ALLEXPORT)) {
+		v->pm->flags |= PMFLAG_x;
+		v->pm->env = addenv(v->pm->nam,val);
+	}
 	switch (pmtype(v->pm)) {
 		case PMFLAG_s:
 			if (v->a == 0 && v->b == -1)
@@ -432,9 +434,8 @@ char *s;
 
 					ad = v->a+1;
 					ss = zalloc((ad+1)*sizeof *ss);
-					memcpy(ss,st,(ad+1)*sizeof *ss);
 					for (t0 = 0; t0 != ac; t0++)
-						ss[t0] = ztrdup(ss[t0]);
+						ss[t0] = ztrdup(st[t0]);
 					while (ac < ad)
 						ss[ac++] = ztrdup("");
 					ss[ac] = NULL;
@@ -454,14 +455,17 @@ char buf[20];
 
 	if (v->pm->flags & PMFLAG_r)
 		return;
+	sprintf(buf,"%ld",val);
 	if (v->pm->env) {
-		sprintf(buf,"%ld",val);
 		v->pm->env = replenv(v->pm->env,buf);
 	}
+	else if (isset(ALLEXPORT)) {
+		v->pm->flags |= PMFLAG_x;
+		v->pm->env = addenv(v->pm->nam,buf);
+	}
 	switch (pmtype(v->pm))
 		{
 		case PMFLAG_s:
-			sprintf(buf,"%ld",val);
 			(v->pm->sets.cfn)(v->pm,ztrdup(buf));
 			break;
 		case PMFLAG_i:
@@ -490,14 +494,64 @@ char buf[20];
 void setarrvalue(v,val) /**/
 Value v;char **val;
 {
-	if (v->pm->flags & PMFLAG_r)
+	if (v->pm->flags & PMFLAG_r) {
+		freearray(val);
 		return;
+	}
 	if (pmtype(v->pm) != PMFLAG_A)
 		{
+		freearray(val);
 		zerr("attempt to assign array value to non-array",NULL,0);
 		return;
 		}
-	(v->pm->sets.afn)(v->pm,val);
+	if (v->a == 0 && v->b == -1)
+		(v->pm->sets.afn)(v->pm,val);
+	else {
+		char **ss = (v->pm->gets.afn)(v->pm);
+		int ac,ad,t0;
+
+		ac = arrlen(ss);
+		if (v->a < 0) {
+			v->a += ac;
+			if (v->a < 0) v->a = 0;
+		}
+		if (v->b < 0) v->b += ac;
+		if (v->b < v->a) v->b = v->a;
+		t0 = arrlen(val) - (v->b - v->a + 1);
+		if (v->b >= ac || t0 != 0) {
+			char **st = ss;
+
+			ad = (v->b > (ac - 1) ? v->b : (ac - 1)) + t0 + 1;
+			ss = zalloc((ad+1)*sizeof *ss);
+			for (t0 = 0; t0 < v->a; t0++)
+				ss[t0] = ztrdup(t0 < ac ? st[t0] : "");
+			while (*val) ss[t0++] = *val++;
+			while (++v->b < ac) ss[t0++] = ztrdup(st[v->b]);
+			ss[t0] = NULL;
+		} else {
+			for (t0 = v->a; t0 <= v->b; t0++) {
+				if (ss[t0]) free(ss[t0]);
+				ss[t0] = val[t0];
+			}
+		}
+		(v->pm->sets.afn)(v->pm,ss);
+	}
+}
+
+int getparamtype(s,l) /**/
+char *s;int l;
+{
+char sav,*t = s;
+Value v;
+
+	sav = t[l];
+	t[l] = '\0';
+
+	if (!(v = getvalue(&s,0)))
+		return -1;
+
+	t[l] = sav;
+	return (pmtype(v->pm));
 }
 
 char *getsparamval(s,l) /**/
@@ -506,8 +560,9 @@ char *s;int l;
 char sav,*t = s;
 Value v;
 
-	if (sav = t[l])
-		t[l] = '\0';
+	sav = t[l];
+	t[l] = '\0';
+	
 	if (!(v = getvalue(&s,0)))
 		return NULL;
 	t[l] = sav;
@@ -540,7 +595,8 @@ char *s;
 {
 Value v;
 
-	if (!(v = getvalue(&s,0))) return NULL;
+	if (!((v = getvalue(&s,0)) && v->isarr))
+		return NULL;
 	return getarrvalue(v);
 }
 
@@ -549,18 +605,27 @@ char *s;char *val;
 {
 Value v;
 char *t = s;
+char *ss;
 
 	if (!isident(s)) {
 		zerr("not an identifier: %s",s,0);
 		free(val);
 		return NULL;
 	}
-	if (!(v = getvalue(&s,1)) || *s)
-		return createparam(t,val,PMFLAG_s);
-	if ((v->pm->flags & PMTYPE) != PMFLAG_s &&
-			!(v->pm->flags & PMFLAG_SPECIAL)) {
-		unsetparam(s);
-		return createparam(t,val,PMFLAG_s);
+	if (ss = strchr(s, '[')) {
+		*ss = '\0';
+		if (!(v = getvalue(&s,1)))
+			createparam(t,zcalloc(sizeof val),PMFLAG_A);
+		*ss = '[';
+		v = getvalue(&t,1);
+	} else {
+		if (!(v = getvalue(&s,1)))
+			return createparam(t,val,PMFLAG_s);
+		if ((v->pm->flags & PMTYPE) != PMFLAG_s &&
+				!(v->pm->flags & PMFLAG_SPECIAL)) {
+			unsetparam(t);
+			return createparam(t,val,PMFLAG_s);
+		}
 	}
 	setstrvalue(v,val);
 	return v->pm;
@@ -571,18 +636,28 @@ char *s;char **val;
 {
 Value v;
 char *t = s;
+char *ss;
 
 	if (!isident(s))
 		{
 		zerr("not an identifier: %s",s,0);
+		freearray(val);
 		return NULL;
 		}
-	if (!(v = getvalue(&s,1)) || *s)
-		return createparam(t,val,PMFLAG_A);
-	if ((v->pm->flags & PMTYPE) != PMFLAG_A &&
-			!(v->pm->flags & PMFLAG_SPECIAL)) {
-		unsetparam(s);
-		return createparam(t,val,PMFLAG_A);
+	if(ss = strchr(s, '[')) {
+		*ss = '\0';
+		if (!(v = getvalue(&s,1)))
+			createparam(t,zcalloc(sizeof val),PMFLAG_A);
+		*ss = '[';
+		v = getvalue(&t,1);
+	} else {
+		if (!(v = getvalue(&s,1)))
+			return createparam(t,val,PMFLAG_A);
+		if ((v->pm->flags & PMTYPE) != PMFLAG_A &&
+				!(v->pm->flags & PMFLAG_SPECIAL)) {
+			unsetparam(t);
+			return createparam(t,val,PMFLAG_A);
+		}
 	}
 	setarrvalue(v,val);
 	return v->pm;
@@ -635,6 +710,8 @@ Param pm;
 	if (pmtype(pm) == PMFLAG_s && (pm->flags & PMFLAG_x)) {
 		delenv(pm->env);
 		free(pm->env);
+		if (pm->flags & PMFLAG_SPECIAL)
+			pm->env = NULL;
 	}
 	if (!(pm->flags & PMFLAG_SPECIAL))
 		freepm(remhnode(s,paramtab));
@@ -760,6 +837,7 @@ char **s,**t,*u,*up;
 			u = up;
 			filesub(&u);
 			if (!*u) u = ".";
+			free(*t);
 			*t = ztrdup(u);
 		}
 	if (pm->data) {
@@ -818,7 +896,7 @@ Param pm;char **x;
 void hostcmdssetfn(pm,x) /**/
 Param pm;char **x;
 {
-	compctl_process(x,CC_HOSTS|CC_FILES,NULL);
+	compctl_process(x,CC_HOSTS,NULL);
 	freearray(x);
 }
 
@@ -988,11 +1066,11 @@ char *dashgetfn(pm) /**/
 Param pm;
 {
 static char buf[100];
-char *val = buf;
+char *val;
 int t0;
 
 	for (val = buf, t0 = ' ';t0 <= 'z'; t0++)
-		if (opts[t0] == OPT_SET)
+		if (isset(t0))
 			*val++ = t0;
 	*val = '\0';
 	return buf;
@@ -1095,12 +1173,13 @@ Param pm;char *x;
 			pp = tbuf;
 			if (tcstr[t0])
 				free(tcstr[t0]);
-			if (!tgetstr(tccapnams[t0],&pp))
+			/* AIX tgetstr() ignores second argument */
+			if (!(pp = tgetstr(tccapnams[t0],&pp)))
 				tcstr[t0] = NULL, tclen[t0] = 0;
 			else
 				{
-				tcstr[t0] = zalloc(tclen[t0] = pp-tbuf);
-				memcpy(tcstr[t0],tbuf,tclen[t0]);
+				tcstr[t0] = zalloc(tclen[t0] = strlen(pp)+1);
+				memcpy(tcstr[t0],pp,tclen[t0]);
 				}
 			}
 
@@ -1156,11 +1235,11 @@ int ct;
 	for (; *envp; envp++,envp2++) {
 		for (str = *envp; *str && *str != '='; str++);
 		if (*str == '=') {
-			char *iname;
+			char *iname = NULL;
 
 			*str = '\0';
-			if (isident(*envp))
-				pm = setsparam(iname = *envp,ztrdup(str+1));
+			pm = (isident(*envp) && !strchr(*envp, '[')) ?
+				setsparam(iname = *envp,ztrdup(str+1)) : NULL;
 			if (pm) {
 				pm->flags |= PMFLAG_x;
 				pm->env = *envp2;
@@ -1311,5 +1390,3 @@ long x;
 	if (v < 0)
 		s[0] = '-';
 }
-
-
