@@ -19,7 +19,6 @@
  */
 
 #include "zsh.h"
-#include "funcs.h"
 
 #define YYERROR { tok = LEXERR; return NULL; }
 #define YYERRORV { tok = LEXERR; return; }
@@ -193,7 +192,7 @@ Pline p,p2;
 		return p;
 	} else if (tok == BARAMP) {
 		struct redir *rdr = alloc(sizeof *rdr);
-		rdr->type = MERGE; rdr->fd1 = 2; rdr->fd2 = 1;
+		rdr->type = MERGEOUT; rdr->fd1 = 2; rdr->fd2 = 1;
 		addnode(c->redir,rdr);
 
 		yylex();
@@ -534,13 +533,13 @@ Cmd c;
 	yylex();
 	c->type = FUNCDEF;
 	c->args = newlist();
+	incmdpos = 1;
 	while (tok == STRING) {
 		if (*tokstr == Inbrace && !tokstr[1]) { tok = INBRACE; break; }
 		addnode(c->args,tokstr);
 		yylex();
 	}
 	nocorrect = 0;
-	incmdpos = 1;
 	if (tok == INOUTPAR) yylex();
 	while (tok == SEPER) yylex();
 	if (tok == INBRACE) {
@@ -645,7 +644,7 @@ int isnull = 1;
 		} else break;
 		isnull = 0;
 	}
-	if (isnull && !full(c->redir)) return NULL;
+	if (isnull && empty(c->redir)) return NULL;
 	if (full(c->args)) {
 		if (underscore)
 			free(underscore);
@@ -795,7 +794,7 @@ int oldcmdpos,oldnc;
 /* <<[-] name */
 
 	} else if (fn->type == HEREDOC || fn->type == HEREDOCDASH) {
-		char tbuf[256], *tlin;
+		char tbuf[256], *tlin = NULL;
 		int tsiz = 0, l;
 		/* Save the rest of the current line for later tokenization */
 		if (!isnewlin) {
