@@ -21,25 +21,8 @@
 #include "zsh.h"
 #include <pwd.h>
 #include <errno.h>
-#ifdef __hpux
-#include <ndir.h>
-#else
-#ifndef SYSV
-#include <sys/dir.h>
-#endif
-#endif
 #include <fcntl.h>
 
-#ifdef SYSV
-#define direct dirent
-#undef TIOCGWINSZ
-int readlink(s,t,z)
-char *s;char *t;int z;
-{
-return -1;
-}
-#endif
-
 /* source a file */
 
 int source(s) /**/
@@ -61,7 +44,7 @@ FILE *obshin = bshin;
 		return 1;
 		}
 	bshin = fdopen(SHIN,"r");
-	loop();
+	loop(0);
 	fclose(bshin);
 	bshin = obshin;
 	opts[SHINSTDIN] = oldshst;
@@ -218,16 +201,6 @@ static struct sigvec vec = { handler,sigmask(SIGINT),0 };
 #endif
 }
 
-char *fgetline(buf,len,in) /**/
-char *buf;int len;FILE *in;
-{
-	if (!fgets(buf,len,in))
-		return NULL;
-	buf[len] = '\0';
-	buf[strlen(buf)-1] = '\0';
-	return buf;
-}
-
 /* get a symlink-free pathname for s relative to PWD */
 
 char *findpwd(s) /**/
@@ -407,13 +380,7 @@ int t0;
 	t0 = finddir(s);
 	if (t0 == -1)
 		{
-		if (!strncmp(s,home,t0 = strlen(home)) && t0 > 1)
-			{
-			putc('~', f);
-			fputs(s+t0,f);
-			}
-		else
-			fputs(s,f);
+		fputs(s,f);
 		}
 	else
 		{
@@ -437,35 +404,45 @@ char *s;
 }
 
 /* see if a path has a named directory as its prefix */
+/* Modified by Carl Edman, cedman@golem.ps.uci.edu.
+   Exhaustive search is more expensive, but binary search just doesn't
+   handle all the cases for directory prefixes very well.
+   Caching improves performance */
 
 int finddir(s) /**/
 char *s;
 {
-int t0,t1,step;
+int t0,t1 = -1;
+int bestlen = -1,len,bestnlen = -1,nlen;
+int hash;
+static int cdir = -1,cval = -1;
 
-	if (userdirsz)
+	if (!s) /* Invalidate directory cache table */
 		{
-		step = t0 = userdirsz/2;
-		for(;;)
+		cval = -1;
+		return -1;
+		}
+	if (!userdirsz) return -1;
+
+	hash = hasher(s);
+
+	if (cval != -1 && cval == hash) return cdir;
+
+	for(t0 = 0; t0 < userdirsz; t0++)
+	    if (userdirs[t0] && !dircmp(userdirs[t0],s))
+		{
+		len = strlen(userdirs[t0]);
+		nlen = strlen(usernames[t0]);
+		if ((len > bestlen) || ((len == bestlen) && (nlen < bestnlen)))
 			{
-			t1 = (userdirs[t0]) ? dircmp(userdirs[t0],s) : 1;
-			if (!t1)
-				{
-				while (t0 != userdirsz-1 && userdirs[t0+1] && 
-						!dircmp(userdirs[t0+1],s)) 
-					t0++;
-				return t0;
-				}
-			if (!step)
-				break;
-			if (t1 > 0)
-				t0 = t0-step+step/2;
-			else
-				t0 += step/2;
-			step /= 2;
+			t1 = t0;
+			bestlen = len;
+			bestnlen = nlen;
 			}
 		}
-	return -1;
+	cval = hash;
+	cdir = t1;
+	return t1;
 }
 
 /* add a named directory */
@@ -481,6 +458,7 @@ int t0,t1;
 		return;
 	if ((t0 = finddir(t)) != -1 && !strcmp(s,usernames[t0]))
 		return;
+	finddir(0); /* Invalidate directory cache table */
 	if (userdirsz == userdirct)
 		{
 		userdirsz *= 2;
@@ -662,7 +640,7 @@ char *v,*u,c;
 					st.st_mtime > lastmailcheck)
 				if (!u)
 					{
-					fprintf(stderr,"You have new mail.\n",*s);
+					fprintf(stderr,"You have new mail.\n");
 					fflush(stderr);
 					}
 				else
@@ -887,28 +865,13 @@ void adjustwinsize() /**/
 int zyztem(s,t) /**/
 char *s;char *t;
 {
-#ifdef WAITPID
-int pid,statusp;
+int cj = thisjob;
 
-	if (!(pid = fork()))
-		{
-		s = tricat(s," ",t);
-		execl("/bin/sh","sh","-c",s,(char *) 0);
-		_exit(1);
-		}
-	while (waitpid(pid,&statusp,WUNTRACED) == -1 && errno == EINTR);
-	if (WIFEXITED(statusp))
-		return WEXITSTATUS(statusp);
-	return 1;
-#else
-	if (!waitfork())
-		{
-		s = tricat(s," ",t);
-		execl("/bin/sh","sh","-c",s,(char *) 0);
-		_exit(1);
-		}
-	return 0;
-#endif
+	s = tricat(s," ",t);
+	execstring(s);	/* Depends on recursion condom in execute() */
+	free(s);
+	thisjob = cj;
+	return lastval;
 }
 
 #ifndef WAITPID
@@ -978,8 +941,8 @@ Cmd c;
 	if (l)
 		{
 		c = l->left->left->left;
-		if (c->type == SIMPLE && !full(c->args) && !full(c->redir)
-				&& !full(c->vars) && !c->flags)
+		if (c->type == SIMPLE && empty(c->args) && empty(c->redir)
+				&& empty(c->vars) && !c->flags)
 			l = NULL;
 		}
 	if (t0 == -1)
@@ -993,16 +956,19 @@ Cmd c;
 	if (!l)
 		{
 		sigtrapped[t0] = 2;
-		if (t0 && t0 < SIGCOUNT && t0 != SIGCHLD)
-			{
-			signal(t0,SIG_IGN);
-			sigtrapped[t0] = 2;
-			}
+		if (t0 && t0 < SIGCOUNT && t0 != SIGCHLD &&
+#ifdef SIGWINCH
+			t0 != SIGWINCH &&
+#endif
+			t0 != SIGHUP) signal(t0,SIG_IGN);
 		}
 	else
 		{
-		if (t0 && t0 < SIGCOUNT && t0 != SIGCHLD)
-			signal(t0,handler);
+		if (t0 && t0 < SIGCOUNT && t0 != SIGCHLD &&
+#ifdef SIGWINCH
+			t0 != SIGWINCH &&
+#endif
+			t0 != SIGHUP) signal(t0,handler);
 		sigtrapped[t0] = 1;
 		permalloc();
 		sigfuncs[t0] = (List) dupstruct(l);
@@ -1021,7 +987,11 @@ int t0;
 	}
 	sigtrapped[t0] = 0;
 	if (t0 == SIGINT) intr();
-	else if (t0 && t0 < SIGCOUNT && t0 != SIGCHLD) signal(t0,SIG_DFL);
+	else if (t0 && t0 < SIGCOUNT && t0 != SIGCHLD &&
+#ifdef SIGWINCH
+		t0 != SIGWINCH &&
+#endif
+		t0 != SIGHUP) signal(t0,SIG_DFL);
 	if (sigfuncs[t0]) freestruct(sigfuncs[t0]);
 }
 
@@ -1040,7 +1010,8 @@ int sav,savval;
 		doshfuncnoval(sigfuncs[sig],NULL,0);
 		lexrestore();
 	}
-	sigtrapped[sig] = sav;
+	if (sigtrapped[sig])
+		sigtrapped[sig] = sav;
 	lastval = savval;
 }
 
@@ -1075,7 +1046,7 @@ char *u = *s;
 	*u = '\0';
 }
 
-void checkrmall(s) /**/
+int checkrmall(s) /**/
 char *s;
 {
 	fflush(stdin);
@@ -1086,7 +1057,7 @@ char *s;
 			(pwd[1]) ? pwd : "",s);
 	fflush(stderr);
 	feep();
-	errflag |= (getquery() != 'y');
+	return (getquery() == 'y');
 }
 
 int getquery() /**/
@@ -1180,7 +1151,7 @@ int pram = 0;
 			if (*guess == Tilde) *guess = '~';
 			else if (*guess == String) *guess = '$';
 			else if (*guess == Equals) *guess = '=';
-			pp = putprompt(sprompt,&junk);
+			pp = putprompt(sprompt,&junk,1);
 			*guess = firstchar;
 			fprintf(stderr,"%s",pp);
 			fflush(stderr);
@@ -1335,11 +1306,9 @@ char *s;
 {
 int ct;
 char *t,**ret,**ptr;
-char **lastmem = NULL;
 
 	for (t = s, ct = 0; *t; t++) if (*t == ':') ct++;
-	if (lastmem) freearray(lastmem);
-	lastmem = ptr = ret = (char **) zalloc(sizeof(char **)*(ct+2));
+	ptr = ret = (char **) zalloc(sizeof(char **)*(ct+2));
 	t = s;
 	do {
 		for (s = t; *t && *t != ':'; t++);
@@ -1386,17 +1355,17 @@ int type;
 int t0;
 struct node *n = (struct node *) alloc(sizeof *n);
 static int typetab[N_COUNT][4] = {
-	NT_NODE,NT_NODE,0,0,
-	NT_NODE,NT_NODE,0,0,
-	NT_NODE,NT_NODE,0,0,
-	NT_STR|NT_LIST,NT_NODE,NT_NODE|NT_LIST,NT_NODE|NT_LIST,
-	NT_STR,0,0,0,
-	NT_NODE,NT_NODE,0,0,
-	NT_STR,NT_NODE,0,0,
-	NT_NODE,NT_STR,NT_NODE,0,
-	NT_NODE,NT_NODE,NT_NODE,0,
-	NT_NODE,NT_NODE,0,0,
-	NT_STR,NT_STR,NT_STR|NT_LIST,0
+	{NT_NODE,NT_NODE,0,0},
+	{NT_NODE,NT_NODE,0,0},
+	{NT_NODE,NT_NODE,0,0},
+	{NT_STR|NT_LIST,NT_NODE,NT_NODE|NT_LIST,NT_NODE|NT_LIST},
+	{NT_STR,0,0,0},
+	{NT_NODE,NT_NODE,0,0},
+	{NT_STR,NT_NODE,0,0},
+	{NT_NODE,NT_STR,NT_NODE,0},
+	{NT_NODE,NT_NODE,NT_NODE,0},
+	{NT_NODE,NT_NODE,0,0},
+	{NT_STR,NT_STR,NT_STR|NT_LIST,0}
 	};
 
 	n->type = type;
@@ -1513,7 +1482,8 @@ Cmd c;
 	if (!l->right)
 		return;
 	if (l->right->right || l->right->left->right ||
-			l->right->left->left->right)
+			l->right->left->flags || l->right->left->left->right ||
+		 	l->left->flags)
 		return;
 	c = l->left->left->left;
 	if (c->type != SIMPLE || full(c->args) || full(c->redir)
@@ -1546,7 +1516,7 @@ char *s;
 	typtab[' '] |= IBLANK|INBLANK;
 	typtab['\t'] |= IBLANK|INBLANK;
 	typtab['\n'] |= INBLANK;
-	for (t0 = (int) (unsigned char) ALPOP; t0 <= (int) (unsigned char) Nularg;
+	for (t0 = (int) STOUC(ALPOP); t0 <= (int) STOUC(Nularg);
 			t0++)
 		typtab[t0] |= ITOK;
 	for (s = ifs; *s; s++)
@@ -1718,12 +1688,12 @@ int gethostname(nameptr, maxlength)
 char *nameptr;
 int maxlength;
 {
-struct utsname *name;
+struct utsname name;
 int result;
 
-	result = uname(name);
+	result = uname(&name);
 	if (result >= 0) {
-		strcpy(nameptr,name->sysname);
+		strncpy(nameptr,name.nodename,maxlength);
 		return 0;
 	} else return -1;
 }
@@ -1791,7 +1761,7 @@ static int ep = 0;
 
 long gettygrp() /**/
 {
-int arg = -1;
+int arg;
 
 	if (SHTTY == -1) return -1;
 #ifdef HAS_TCSETPGRP
