@@ -19,10 +19,10 @@
  */
 
 #include "zsh.h"
-#include <utmp.h>
 
 static int wtabsz;
 struct utmp *wtab;
+static time_t lastutmpcheck;
 
 /* get the time of login/logout for WATCH */
 
@@ -65,7 +65,7 @@ int srchlimit = 50; /* max number of wtmp records to search */
 			fclose(in);
 			return time(NULL);
 			}
-	while (strncmp(uu.ut_line,u->ut_line,8));
+	while (strncmp(uu.ut_line,u->ut_line,sizeof(u->ut_line)));
 	fclose(in);
 	return uu.ut_time;
 }
@@ -78,7 +78,7 @@ int inout;struct utmp *u;char *fmt;
 char *p,buf[40],*bf;
 int i;
 time_t timet;
-struct tm *tm = NULL;
+struct tm *tm;
 
 	while (*fmt)
 		if (*fmt != '%')
@@ -89,20 +89,20 @@ struct tm *tm = NULL;
 			switch(*fmt++)
 				{
 				case 'n':
-					printf("%.*s",8,u->ut_name);
+					printf("%.*s",sizeof(u->ut_name),u->ut_name);
 					break;
 				case 'a':
 					printf("%s",(!inout) ? "logged off" : "logged on");
 					break;
 				case 'l':
-					if (u->ut_line[0] == 't')
-						printf("%.*s",5,u->ut_line+3);
+					if (!strncmp(u->ut_line, "tty", 3))
+						printf("%.*s",sizeof(u->ut_line)-3,u->ut_line+3);
 					else
-						printf("%.*s",8,u->ut_line);
+						printf("%.*s",sizeof(u->ut_line),u->ut_line);
 					break;
 #ifdef UTMP_HOST
 				case 'm':
-					for (p = u->ut_host,i = 16; i && *p;i--,p++)
+					for (p = u->ut_host,i = sizeof(u->ut_host); i && *p;i--,p++)
 						{
 						if (*p == '.' && !idigit(p[1]))
 							break;
@@ -110,7 +110,7 @@ struct tm *tm = NULL;
 						}
 					break;
 				case 'M':
-					printf("%.*s",16,u->ut_host);
+					printf("%.*s",sizeof(u->ut_host),u->ut_host);
 					break;
 #endif
 				case 't':
@@ -199,6 +199,12 @@ int bad;
 		watchlog2(inout,u,fmt);
 		return;
 		}
+	if (*w && !strcmp(*w,"notme") &&
+	  strncmp(u->ut_name, username, sizeof(u->ut_name)))
+		{
+		watchlog2(inout,u,fmt);
+		return;
+		}
 	for (; *w; w++)
 		{
 		bad = 0;
@@ -208,7 +214,7 @@ int bad;
 			for (vv = v; *vv && *vv != '@' && *vv != '%'; vv++);
 			sav = *vv;
 			*vv = '\0';
-			if (strncmp(u->ut_name,v,8))
+			if (strncmp(u->ut_name,v,sizeof(u->ut_name)))
 				bad = 1;
 			*vv = sav;
 			v = vv;
@@ -219,7 +225,7 @@ int bad;
 				for (vv = ++v; *vv && *vv != '@'; vv++);
 				sav = *vv;
 				*vv = '\0';
-				if (strncmp(u->ut_line,v,8))
+				if (strncmp(u->ut_line,v,sizeof(u->ut_line)))
 					bad = 1;
 				*vv = sav;
 				v = vv;
@@ -252,7 +258,7 @@ int ucmp(u,v) /**/
 struct utmp *u;struct utmp *v;
 {
 	if (u->ut_time == v->ut_time)
-		return strncmp(u->ut_line,v->ut_line,8);
+		return strncmp(u->ut_line,v->ut_line,sizeof(u->ut_line));
 	return u->ut_time - v->ut_time;
 }
 
@@ -281,7 +287,8 @@ FILE *in;
 			}
 	fclose(in);
 	if (wtabsz)
-		qsort(wtab,wtabsz,sizeof(struct utmp),ucmp);
+		qsort((vptr)wtab,wtabsz,sizeof(struct utmp),
+				(int (*) DCLPROTO((const void *, const void *)))ucmp);
 }
 
 /* check for login/logout events; executed before each prompt
@@ -294,6 +301,7 @@ char *fmt = (watchfmt) ? watchfmt : DEFWATCHFMT;
 FILE *in;
 int utabsz = 0,utabmax = wtabsz+4,uct,wct;
 struct utmp *utab,*uptr,*wptr;
+struct stat st;
 
 	holdintr();
 	if (!fmt)
@@ -303,6 +311,9 @@ struct utmp *utab,*uptr,*wptr;
 		noholdintr();
 		return;
 	}
+	if ((stat(UTMP_FILE,&st) == -1) || (st.st_mtime <= lastutmpcheck))
+		return;
+	lastutmpcheck = st.st_mtime;
 	uptr = utab = (struct utmp *) zalloc(utabmax*sizeof(struct utmp));
 	if (!(in = fopen(UTMP_FILE,"r"))) {
 		free(utab);
@@ -327,7 +338,8 @@ struct utmp *utab,*uptr,*wptr;
 		return;
 	}
 	if (utabsz)
-		qsort(utab,utabsz,sizeof(struct utmp),ucmp);
+		qsort((vptr)utab,utabsz,sizeof(struct utmp),
+				(int (*) DCLPROTO((const void *, const void *)))ucmp);
 
 	wct = wtabsz; uct = utabsz;
 	uptr = utab; wptr = wtab;
@@ -357,6 +369,7 @@ char *nam;char **argv;char *ops;int func;
 		free(wtab);
 	wtab = (struct utmp *) zalloc(1);
 	wtabsz = 0;
+	lastutmpcheck = 0;
 	dowatch();
 	return 0;
 }
