@@ -23,154 +23,156 @@
 
 
 struct zlecmd zlecmds[] = {
-"accept-and-hold",acceptandhold,0,
-"accept-and-infer-next-history",acceptandinfernexthistory,0,
-"accept-and-menu-complete", acceptandmenucomplete, ZLE_MENUCMP,
-"accept-line",acceptline,0,
-"accept-line-and-down-history",acceptlineanddownhistory,0,
-"backward-char",backwardchar,ZLE_MOVEMENT,
-"backward-delete-char",backwarddeletechar,0,
-"backward-delete-word",backwarddeleteword,0,
-"backward-kill-line",backwardkillline,ZLE_KILL,
-"backward-kill-word",backwardkillword,ZLE_KILL,
-"backward-word",backwardword,ZLE_MOVEMENT,
-"beginning-of-buffer-or-history",beginningofbufferorhistory,ZLE_MOVEMENT,
-"beginning-of-history",beginningofhistory,0,
-"beginning-of-line",beginningofline,ZLE_MOVEMENT,
-"beginning-of-line-hist",beginningoflinehist,ZLE_MOVEMENT,
-"capitalize-word",capitalizeword,0,
-"clear-screen",clearscreen,0,
-"complete-word",completeword,ZLE_MENUCMP,
-"copy-prev-word",copyprevword,0,
-"copy-region-as-kill",copyregionaskill,ZLE_KILL,
-"delete-char",deletechar,0,
-"delete-char-or-list",deletecharorlist,ZLE_MENUCMP,
-"delete-word",deleteword,0,
-"digit-argument",digitargument,ZLE_ARG,
-"down-case-word",downcaseword,0,
-"down-history",downhistory,0,
-"down-line-or-history",downlineorhistory,ZLE_MOVEMENT|ZLE_LINEMOVE,
-"end-of-buffer-or-history",endofbufferorhistory,ZLE_MOVEMENT,
-"end-of-history",endofhistory,0,
-"end-of-line",endofline,ZLE_MOVEMENT,
-"end-of-line-hist",endoflinehist,ZLE_MOVEMENT,
-"exchange-point-and-mark",exchangepointandmark,ZLE_MOVEMENT,
-"execute-last-named-cmd",NULL,0,
-"execute-named-cmd",NULL,0,
-"expand-history",expandhistory,0,
-"expand-or-complete",expandorcomplete,ZLE_MENUCMP,
-"expand-word",expandword,0,
-"forward-char",forwardchar,ZLE_MOVEMENT,
-"forward-word",forwardword,ZLE_MOVEMENT,
-"get-line",getline,0,
-"gosmacs-transpose-chars",gosmacstransposechars,0,
-"history-incremental-search-backward",historyincrementalsearchbackward,0,
-"history-incremental-search-forward",historyincrementalsearchforward,0,
-"history-search-backward",historysearchbackward,ZLE_HISTSEARCH,
-"history-search-forward",historysearchforward,ZLE_HISTSEARCH,
-"infer-next-history",infernexthistory,0,
-"insert-last-word",insertlastword,ZLE_INSERT,
-"kill-buffer",killbuffer,ZLE_KILL,
-"kill-line",killline,ZLE_KILL,
-"kill-region",killregion,ZLE_KILL,
-"kill-whole-line",killwholeline,ZLE_KILL,
-"list-choices",listchoices,0,
-"list-expand",listexpand,ZLE_MENUCMP,
-"magic-space",magicspace,0,
-"menu-complete",menucompleteword,ZLE_MENUCMP,
-"menu-expand-or-complete",menuexpandorcomplete,ZLE_MENUCMP,
-"overwrite-mode",overwritemode,0,
-"push-line",pushline,0,
-"quoted-insert",quotedinsert,ZLE_INSERT,
-"quote-line",quoteline,0,
-"quote-region",quoteregion,0,
-"redisplay",redisplay,0,
-"reverse-menu-complete",reversemenucomplete,ZLE_MENUCMP,
-"run-help",processcmd,0,
-"self-insert",selfinsert,ZLE_INSERT,
-"self-insert-unmeta",selfinsertunmeta,ZLE_INSERT,
-"send-break",sendbreak,0,
-"send-string",sendstring,0,
-"",NULL,0,
-"set-mark-command",setmarkcommand,0,
-"spell-word",spellword,0,
-"toggle-literal-history",toggleliteralhistory,0,
-"transpose-chars",transposechars,0,
-"transpose-words",transposewords,0,
-"undefined-key",undefinedkey,0,
-"undo",undo,ZLE_UNDO,
-"universal-argument",universalargument,ZLE_ARG,
-"up-case-word",upcaseword,0,
-"up-history",uphistory,0,
-"up-line-or-history",uplineorhistory,ZLE_LINEMOVE|ZLE_MOVEMENT,
-"vi-add-eol",viaddeol,0,
-"vi-add-next",viaddnext,0,
-"vi-backward-blank-word",vibackwardblankword,ZLE_MOVEMENT,
-"vi-backward-char",vibackwardchar,ZLE_MOVEMENT,
-"vi-backward-delete-char",vibackwarddeletechar,ZLE_KILL,
-"vi-beginning-of-line",vibeginningofline,ZLE_MOVEMENT,
-"vi-caps-lock-panic",vicapslockpanic,0,
-"vi-change",vichange,0,
-"vi-change-eol",vichangeeol,0,
-"vi-change-whole-line",vichangewholeline,0,
-"vi-cmd-mode",vicmdmode,0,
-"vi-delete",videlete,ZLE_KILL,
-"vi-delete-char",videletechar,ZLE_KILL,
-"vi-digit-or-beginning-of-line",NULL,0,
-"vi-end-of-line",viendofline,ZLE_MOVEMENT,
-"vi-fetch-history",vifetchhistory,0,
-"vi-find-next-char",vifindnextchar,ZLE_MOVEMENT,
-"vi-find-next-char-skip",vifindnextcharskip,ZLE_MOVEMENT,
-"vi-find-prev-char",vifindprevchar,ZLE_MOVEMENT,
-"vi-find-prev-char-skip",vifindprevcharskip,ZLE_MOVEMENT,
-"vi-first-non-blank",vifirstnonblank,ZLE_MOVEMENT,
-"vi-forward-blank-word",viforwardblankword,ZLE_MOVEMENT,
-"vi-forward-blank-word-end",viforwardblankwordend,ZLE_MOVEMENT,
-"vi-forward-char",viforwardchar,ZLE_MOVEMENT,
-"vi-forward-word-end",viforwardwordend,ZLE_MOVEMENT,
-"vi-goto-column",vigotocolumn,ZLE_MOVEMENT,
-"vi-goto-mark",vigotomark,ZLE_MOVEMENT,
-"vi-goto-mark-line",vigotomarkline,ZLE_MOVEMENT,
-"vi-history-search-backward",vihistorysearchbackward,0,
-"vi-history-search-forward",vihistorysearchforward,0,
-"vi-indent",viindent,0,
-"vi-insert",viinsert,0,
-"vi-insert-bol",viinsertbol,0,
-"vi-join",vijoin,0,
-"vi-match-bracket",vimatchbracket,ZLE_MOVEMENT,
-"vi-open-line-above",viopenlineabove,0,
-"vi-open-line-below",viopenlinebelow,0,
-"vi-oper-swap-case",vioperswapcase,0,
-"vi-put-after",viputafter,ZLE_YANK,
-"vi-repeat-change",virepeatchange,0,
-"vi-repeat-find",virepeatfind,ZLE_MOVEMENT,
-"vi-repeat-search",virepeatsearch,ZLE_MOVEMENT,
-"vi-replace",vireplace,0,
-"vi-replace-chars",vireplacechars,0,
-"vi-rev-repeat-find",virevrepeatfind,ZLE_MOVEMENT,
-"vi-rev-repeat-search",virevrepeatsearch,ZLE_MOVEMENT,
-"vi-set-buffer",visetbuffer,0,
-"vi-set-mark",visetmark,0,
-"vi-substitute",visubstitute,0,
-"vi-swap-case",viswapcase,0,
-"vi-undo-change",undo,0,
-"vi-unindent",viunindent,0,
-"vi-yank",viyank,0,
-"vi-yank-eol",viyankeol,0,
-"which-command",processcmd,0,
-"yank",yank,ZLE_YANK|ZLE_NAMEDBUFFER,
-"yank-pop",yankpop,ZLE_YANK,
-"emacs-forward-word",emacsforwardword,ZLE_MOVEMENT,
-"emacs-backward-word",emacsbackwardword,ZLE_MOVEMENT,
-"kill-word",killword,ZLE_KILL,
-"vi-kill-line",vikillline,0,
-"vi-backward-kill-word",vibackwardkillword,0,
-"expand-cmd-path",expandcmdpath,0,
-"neg-argument",negargument,ZLE_NEGARG,
-"pound-insert",poundinsert,0,
-"vi-forward-word",viforwardword,ZLE_MOVEMENT,
-"vi-backward-word",vibackwardword,ZLE_MOVEMENT,
-"",NULL,0
+{"accept-and-hold",acceptandhold,0},
+{"accept-and-infer-next-history",acceptandinfernexthistory,0},
+{"accept-and-menu-complete", acceptandmenucomplete, ZLE_MENUCMP},
+{"accept-line",acceptline,0},
+{"accept-line-and-down-history",acceptlineanddownhistory,0},
+{"backward-char",backwardchar,ZLE_MOVEMENT},
+{"backward-delete-char",backwarddeletechar,ZLE_DELETE},
+{"backward-delete-word",backwarddeleteword,ZLE_DELETE},
+{"backward-kill-line",backwardkillline,ZLE_KILL},
+{"backward-kill-word",backwardkillword,ZLE_KILL|ZLE_DELETE},
+{"backward-word",backwardword,ZLE_MOVEMENT},
+{"beginning-of-buffer-or-history",beginningofbufferorhistory,ZLE_MOVEMENT},
+{"beginning-of-history",beginningofhistory,0},
+{"beginning-of-line",beginningofline,ZLE_MOVEMENT},
+{"beginning-of-line-hist",beginningoflinehist,ZLE_MOVEMENT},
+{"capitalize-word",capitalizeword,0},
+{"clear-screen",clearscreen,0},
+{"complete-word",completeword,ZLE_MENUCMP},
+{"copy-prev-word",copyprevword,0},
+{"copy-region-as-kill",copyregionaskill,ZLE_KILL},
+{"delete-char",deletechar,ZLE_DELETE},
+{"delete-char-or-list",deletecharorlist,ZLE_MENUCMP},
+{"delete-word",deleteword,ZLE_DELETE},
+{"digit-argument",digitargument,ZLE_ARG},
+{"down-case-word",downcaseword,0},
+{"down-history",downhistory,0},
+{"down-line-or-history",downlineorhistory,ZLE_MOVEMENT|ZLE_LINEMOVE},
+{"end-of-buffer-or-history",endofbufferorhistory,ZLE_MOVEMENT},
+{"end-of-history",endofhistory,0},
+{"end-of-line",endofline,ZLE_MOVEMENT},
+{"end-of-line-hist",endoflinehist,ZLE_MOVEMENT},
+{"exchange-point-and-mark",exchangepointandmark,ZLE_MOVEMENT},
+{"execute-last-named-cmd",(F) 0,0},
+{"execute-named-cmd",(F) 0,0},
+{"expand-history",expandhistory,0},
+{"expand-or-complete",expandorcomplete,ZLE_MENUCMP},
+{"expand-word",expandword,0},
+{"forward-char",forwardchar,ZLE_MOVEMENT},
+{"forward-word",forwardword,ZLE_MOVEMENT},
+{"get-line",getline,0},
+{"gosmacs-transpose-chars",gosmacstransposechars,0},
+{"history-incremental-search-backward",historyincrementalsearchbackward,0},
+{"history-incremental-search-forward",historyincrementalsearchforward,0},
+{"history-search-backward",historysearchbackward,ZLE_HISTSEARCH},
+{"history-search-forward",historysearchforward,ZLE_HISTSEARCH},
+{"infer-next-history",infernexthistory,0},
+{"insert-last-word",insertlastword,ZLE_INSERT},
+{"kill-buffer",killbuffer,ZLE_KILL},
+{"kill-line",killline,ZLE_KILL},
+{"kill-region",killregion,ZLE_KILL},
+{"kill-whole-line",killwholeline,ZLE_KILL},
+{"list-choices",listchoices,ZLE_DELETE},	/* ZLE_DELETE fixes autoremoveslash */
+{"list-expand",listexpand,ZLE_MENUCMP},
+{"magic-space",magicspace,0},
+{"menu-complete",menucompleteword,ZLE_MENUCMP},
+{"menu-expand-or-complete",menuexpandorcomplete,ZLE_MENUCMP},
+{"overwrite-mode",overwritemode,0},
+{"push-line",pushline,0},
+{"quoted-insert",quotedinsert,ZLE_INSERT},
+{"quote-line",quoteline,0},
+{"quote-region",quoteregion,0},
+{"redisplay",redisplay,0},
+{"reverse-menu-complete",reversemenucomplete,ZLE_MENUCMP},
+{"run-help",processcmd,0},
+{"self-insert",selfinsert,ZLE_INSERT},
+{"self-insert-unmeta",selfinsertunmeta,ZLE_INSERT},
+{"send-break",sendbreak,0},
+{"send-string",sendstring,0},
+{"",(F) 0,0},
+{"set-mark-command",setmarkcommand,0},
+{"spell-word",spellword,0},
+{"toggle-literal-history",toggleliteralhistory,0},
+{"transpose-chars",transposechars,0},
+{"transpose-words",transposewords,0},
+{"undefined-key",undefinedkey,0},
+{"undo",undo,ZLE_UNDO},
+{"universal-argument",universalargument,ZLE_ARG},
+{"up-case-word",upcaseword,0},
+{"up-history",uphistory,0},
+{"up-line-or-history",uplineorhistory,ZLE_LINEMOVE|ZLE_MOVEMENT},
+{"vi-add-eol",viaddeol,0},
+{"vi-add-next",viaddnext,0},
+{"vi-backward-blank-word",vibackwardblankword,ZLE_MOVEMENT},
+{"vi-backward-char",vibackwardchar,ZLE_MOVEMENT},
+{"vi-backward-delete-char",vibackwarddeletechar,ZLE_KILL},
+{"vi-beginning-of-line",vibeginningofline,ZLE_MOVEMENT},
+{"vi-caps-lock-panic",vicapslockpanic,0},
+{"vi-change",vichange,0},
+{"vi-change-eol",vichangeeol,0},
+{"vi-change-whole-line",vichangewholeline,0},
+{"vi-cmd-mode",vicmdmode,0},
+{"vi-delete",videlete,ZLE_KILL},
+{"vi-delete-char",videletechar,ZLE_KILL},
+{"vi-digit-or-beginning-of-line",(F) 0,0},
+{"vi-end-of-line",viendofline,ZLE_MOVEMENT},
+{"vi-fetch-history",vifetchhistory,0},
+{"vi-find-next-char",vifindnextchar,ZLE_MOVEMENT},
+{"vi-find-next-char-skip",vifindnextcharskip,ZLE_MOVEMENT},
+{"vi-find-prev-char",vifindprevchar,ZLE_MOVEMENT},
+{"vi-find-prev-char-skip",vifindprevcharskip,ZLE_MOVEMENT},
+{"vi-first-non-blank",vifirstnonblank,ZLE_MOVEMENT},
+{"vi-forward-blank-word",viforwardblankword,ZLE_MOVEMENT},
+{"vi-forward-blank-word-end",viforwardblankwordend,ZLE_MOVEMENT},
+{"vi-forward-char",viforwardchar,ZLE_MOVEMENT},
+{"vi-forward-word-end",viforwardwordend,ZLE_MOVEMENT},
+{"vi-goto-column",vigotocolumn,ZLE_MOVEMENT},
+{"vi-goto-mark",vigotomark,ZLE_MOVEMENT},
+{"vi-goto-mark-line",vigotomarkline,ZLE_MOVEMENT},
+{"vi-history-search-backward",vihistorysearchbackward,0},
+{"vi-history-search-forward",vihistorysearchforward,0},
+{"vi-indent",viindent,0},
+{"vi-insert",viinsert,0},
+{"vi-insert-bol",viinsertbol,0},
+{"vi-join",vijoin,0},
+{"vi-match-bracket",vimatchbracket,ZLE_MOVEMENT},
+{"vi-open-line-above",viopenlineabove,0},
+{"vi-open-line-below",viopenlinebelow,0},
+{"vi-oper-swap-case",vioperswapcase,0},
+{"vi-put-after",viputafter,ZLE_YANK},
+{"vi-repeat-change",virepeatchange,ZLE_ARG},
+{"vi-repeat-find",virepeatfind,ZLE_MOVEMENT},
+{"vi-repeat-search",virepeatsearch,ZLE_MOVEMENT},
+{"vi-replace",vireplace,0},
+{"vi-replace-chars",vireplacechars,0},
+{"vi-rev-repeat-find",virevrepeatfind,ZLE_MOVEMENT},
+{"vi-rev-repeat-search",virevrepeatsearch,ZLE_MOVEMENT},
+{"vi-set-buffer",visetbuffer,0},
+{"vi-set-mark",visetmark,0},
+{"vi-substitute",visubstitute,0},
+{"vi-swap-case",viswapcase,0},
+{"vi-undo-change",undo,0},
+{"vi-unindent",viunindent,0},
+{"vi-yank",viyank,0},
+{"vi-yank-eol",viyankeol,0},
+{"which-command",processcmd,0},
+{"yank",yank,ZLE_YANK|ZLE_NAMEDBUFFER},
+{"yank-pop",yankpop,ZLE_YANK},
+{"emacs-forward-word",emacsforwardword,ZLE_MOVEMENT},
+{"emacs-backward-word",emacsbackwardword,ZLE_MOVEMENT},
+{"kill-word",killword,ZLE_KILL},
+{"vi-kill-line",vikillline,0},
+{"vi-backward-kill-word",vibackwardkillword,ZLE_KILL},
+{"expand-cmd-path",expandcmdpath,0},
+{"neg-argument",negargument,ZLE_NEGARG|ZLE_ARG},
+{"pound-insert",poundinsert,0},
+{"vi-forward-word",viforwardword,ZLE_MOVEMENT},
+{"vi-backward-word",vibackwardword,ZLE_MOVEMENT},
+{"up-line-or-search",uplineorsearch,ZLE_MOVEMENT|ZLE_LINEMOVE|ZLE_HISTSEARCH},
+{"down-line-or-search",downlineorsearch,ZLE_MOVEMENT|ZLE_LINEMOVE|ZLE_HISTSEARCH},
+{"",(F) 0,0}
 };
 
 int emacsbind[256] = {
