@@ -29,13 +29,13 @@ char *s;
 		{
 		if (curhistline)
 			free(curhistline);
-		curhistline = ztrdup(line);
+		curhistline = ztrdup(UTOSCP(line));
 		}
 	lithist ^= 1;
 	if (!(s = qgetevent(histline)))
 		feep();
 	else
-		sethistline(s);
+		sethistline(STOUCP(s));
 }
 
 void uphistory() /**/
@@ -47,7 +47,7 @@ char *s;
 		{
 		if (curhistline)
 			free(curhistline);
-		curhistline = ztrdup(line);
+		curhistline = ztrdup(UTOSCP(line));
 		}
 	histline -= mult;
 	if (!(s = qgetevent(histline)))
@@ -56,7 +56,7 @@ char *s;
 		histline += mult;
 		}
 	else
-		sethistline(s);
+		sethistline(STOUCP(s));
 }
 
 void uplineorhistory() /**/
@@ -88,6 +88,35 @@ int ocs = cs;
 	}
 }
 
+void uplineorsearch() /**/
+{
+int ocs = cs;
+
+	if (mult < 0) { mult = -mult; downlineorsearch(); return; }
+	if ((lastcmd & ZLE_LINEMOVE) != ZLE_LINEMOVE)
+		lastcol = cs-findbol();
+	cs = findbol();
+	while (mult) {
+		if (!cs)
+			break;
+		cs--;
+		cs = findbol();
+		mult--;
+	}
+	if (mult) {
+		cs = ocs;
+		if (virangeflag) {
+			feep();
+			return;
+		}
+		historysearchbackward();
+	} else {
+		int x = findeol();
+		if ((cs += lastcol) > x)
+			cs = x;
+	}
+}
+
 void downlineorhistory() /**/
 {
 int ocs = cs;
@@ -116,6 +145,34 @@ int ocs = cs;
 	}
 }
 
+void downlineorsearch() /**/
+{
+int ocs = cs;
+
+	if (mult < 0) { mult = -mult; uplineorsearch(); return; }
+	if ((lastcmd & ZLE_LINEMOVE) != ZLE_LINEMOVE)
+		lastcol = cs-findbol();
+	while (mult) {
+		int x = findeol();
+		if (x == ll)
+			break;
+		cs = x+1;
+		mult--;
+	}
+	if (mult) {
+		cs = ocs;
+		if (virangeflag) {
+			feep();
+			return;
+		}
+		historysearchforward();
+	} else {
+		int x = findeol();
+		if ((cs += lastcol) > x)
+			cs = x;
+	}
+}
+
 void acceptlineanddownhistory() /**/
 {
 char *s,*t;
@@ -145,7 +202,7 @@ char *s;
 		histline -= mult;
 		return;
 		}
-	sethistline(s);
+	sethistline(STOUCP(s));
 }
 
 static int histpos;
@@ -159,7 +216,7 @@ char *s;
 		{
 		if (curhistline)
 			free(curhistline);
-		curhistline = ztrdup(line);
+		curhistline = ztrdup(UTOSCP(line));
 		}
 	if (lastcmd & ZLE_HISTSEARCH) t0 = histpos;
 	else for (t0 = 0; line[t0] && iword(line[t0]); t0++);
@@ -173,10 +230,10 @@ char *s;
 			histline = ohistline;
 			return;
 			}
-		if (!hstrncmp(s,line,t0) && hstrcmp(s,line))
+		if (!hstrncmp(s,UTOSCP(line),t0) && hstrcmp(s,UTOSCP(line)))
 			break;
 		}
-	sethistline(s);
+	sethistline(STOUCP(s));
 }
 
 void historysearchforward() /**/
@@ -188,7 +245,7 @@ char *s;
 		{
 		if (curhistline)
 			free(curhistline);
-		curhistline = ztrdup(line);
+		curhistline = ztrdup(UTOSCP(line));
 		}
 	if (lastcmd & ZLE_HISTSEARCH) t0 = histpos;
 	else for (t0 = 0; line[t0] && iword(line[t0]); t0++);
@@ -202,10 +259,10 @@ char *s;
 			histline = ohistline;
 			return;
 			}
-		if (!hstrncmp(s,line,t0) && hstrcmp(s,line))
+		if (!hstrncmp(s,UTOSCP(line),t0) && hstrcmp(s,UTOSCP(line)))
 			break;
 		}
-	sethistline(s);
+	sethistline(STOUCP(s));
 }
 
 void beginningofbufferorhistory() /**/
@@ -224,7 +281,7 @@ char *s;
 		{
 		if (curhistline)
 			free(curhistline);
-		curhistline = ztrdup(line);
+		curhistline = ztrdup(UTOSCP(line));
 		}
 	if (!(s = qgetevent(firsthist())))
 		{
@@ -232,7 +289,7 @@ char *s;
 		return;
 		}
 	histline = firsthist();
-	sethistline(s);
+	sethistline(STOUCP(s));
 }
 
 void endofbufferorhistory() /**/
@@ -250,7 +307,7 @@ void endofhistory() /**/
 	} else
 		{
 		histline = curhist;
-		sethistline(curhistline);
+		sethistline(STOUCP(curhistline));
 		}
 }
 
@@ -307,7 +364,7 @@ int ev;
 void pushline() /**/
 {
 	if (mult < 0) return;
-	pushnode(bufstack,ztrdup(line));
+	pushnode(bufstack,ztrdup(UTOSCP(line)));
 	while (--mult)
 		pushnode(bufstack,ztrdup(""));
 	stackcs = cs;
@@ -349,16 +406,16 @@ int dir;
 char *s,*oldl;
 char ibuf[256],*sbuf = ibuf+10;
 int sbptr = 0,ch,ohl = histline,ocs = cs;
-int nomatch = 0,chequiv = 0;
+int nomatch,chequiv = 0;
 
 	strcpy(ibuf,"i-search: ");
 	statusline = ibuf;
-	oldl = ztrdup(line);
+	oldl = ztrdup(UTOSCP(line));
 	if (histline == curhist)
 		{
 		if (curhistline)
 			free(curhistline);
-		curhistline = ztrdup(line);
+		curhistline = ztrdup(UTOSCP(line));
 		}
 	for (;;)
 		{
@@ -381,9 +438,9 @@ int nomatch = 0,chequiv = 0;
 				if ((sbuf[0] == '^') ?
 						(t = (hstrncmp(s,sbuf+1,sbptr-1)) ? NULL : s) :
 						(t = hstrnstr(s,sbuf,sbptr)))
-					if (!(chequiv && !hstrcmp(line,s)))
+					if (!(chequiv && !hstrcmp(UTOSCP(line),s)))
 						{
-						sethistline(s);
+						sethistline(STOUCP(s));
 						cs = t-s+sbptr-(sbuf[0] == '^');
 						break;
 						}
@@ -456,7 +513,7 @@ char *s,*t;
 		{
 		if (!(s = qgetevent(t0)))
 			return;
-		if (!hstrncmp(s,line,ll))
+		if (!hstrncmp(s,UTOSCP(line),ll))
 			break;
 		}
 	if (!(s = qgetevent(t0+1)))
@@ -494,7 +551,7 @@ char *s,*t;
 		return;
 		}
 	histline = t0+1;
-	sethistline(s);
+	sethistline(STOUCP(s));
 }
 
 void vifetchhistory() /**/
@@ -510,20 +567,20 @@ char *s;
 		}
 		if (curhistline)
 			free(curhistline);
-		curhistline = ztrdup(line);
+		curhistline = ztrdup(UTOSCP(line));
 	}
 	if (!(lastcmd & ZLE_ARG)) mult = curhist;
 	if (!(s = qgetevent(mult)))
 		feep();
 	else {
 		histline = mult;
-		sethistline(s);
+		sethistline(STOUCP(s));
 	}
 }
 
 int getvisrchstr() /**/
 {
-char sbuf[80];
+static char sbuf[80];
 int sptr = 1;
 
 	if (visrchstr)
@@ -586,7 +643,7 @@ char *s;
 		{
 		if (curhistline)
 			free(curhistline);
-		curhistline = ztrdup(line);
+		curhistline = ztrdup(UTOSCP(line));
 		}
 	for (;;)
 		{
@@ -597,7 +654,7 @@ char *s;
 			histline = ohistline;
 			return;
 			}
-		if (!hstrcmp(line,s))
+		if (!hstrcmp(UTOSCP(line),s))
 			continue;
 		if (*visrchstr == '^')
 			{
@@ -608,7 +665,7 @@ char *s;
 			if (hstrnstr(s,visrchstr,t0))
 				break;
 		}
-	sethistline(s);
+	sethistline(STOUCP(s));
 }
 
 void virevrepeatsearch() /**/
