@@ -34,9 +34,11 @@ static Key cky;
 void setterm() /**/
 {
 struct ttyinfo ti;
+#ifdef CLOBBERS_TYPEAHEAD
 #ifdef FIONREAD
 long val;
 #endif
+#endif
 
 #ifdef CLOBBERS_TYPEAHEAD
 #ifdef FIONREAD
@@ -107,24 +109,36 @@ char *s;int len;
 		ungetkey(*--s);
 }
 
+#if (defined(hpux) || defined(pyr)) && defined(HAS_SELECT)
+static int
+breakread(fd, buf, n)
+int fd, n;
+char *buf;
+{
+	fd_set	f;
+
+	FD_ZERO(&f);
+	FD_SET(fd, &f);
+	return(select(fd+1, &f, NULL, NULL, NULL) == -1 ? -1 : read(fd, buf, n));
+}
+#define	read	breakread
+#endif
+
 unsigned int getkey(tmok) /**/
 int tmok;
 {
 char cc;
-unsigned int ret = 0;
-int die = 0;
-#ifdef HAS_SELECT
-fd_set foofd;
-#endif
+unsigned int ret;
+int die = 0, r;
 
 	if (kungetct)
 		ret = (unsigned int) (unsigned char) kungetbuf[--kungetct];
 	else {
-		while (
-#ifdef HAS_SELECT
-		FD_SET(0,&foofd), select(1,&foofd,NULL,NULL,NULL) != 1 ||
-#endif
-		read(0,&cc,1) != 1)
+		while ( (r = read(0,&cc,1)) != 1){
+			if ( r == 0 ){
+				stopmsg = 1;
+				zexit(1);
+			}
 			if (errno == EINTR) {
 				if (!errflag)
 					continue;
@@ -136,15 +150,17 @@ fd_set foofd;
 				fcntl(0,F_SETFL,0);
 			} else if (errno == EIO && !die) {
 				ret = jobbing;
-				jobbing = 1;
+				jobbingv = 1;
 				attachtty(mypgrp);
 				refresh(); /* kludge! */
-				jobbing = ret;
+				jobbingv = ret;
 				die = 1;
-			} else {
+			} else if (errno != 0) {
 				zerr("error on TTY read: %e",NULL,errno);
-				exit(1);
+				stopmsg = 1;
+				zexit(1);
 			}
+		}
 		ret = (unsigned int) (unsigned char) cc;
 	}
 	if (vichgflag) {
@@ -217,7 +233,7 @@ fd_set foofd;
 	if (tmout)
 		alarm(tmout);
 	refresh();
-	errflag = 0;
+	errflag = retflag = 0;
 	while (!done && !errflag)
 		{
 		struct zlecmd *zc;
@@ -238,7 +254,7 @@ fd_set foofd;
 				addundo();
 			if (!(zc->flags & ZLE_MENUCMP)) {
 				if (menucmp) freemenu();
-				if (addedslash &&
+				if (addedslash && !(zc->flags & ZLE_DELETE) &&
 						!((zc->flags & ZLE_INSERT) && c != ' ')) {
 					backdel(1);
 				}
@@ -380,41 +396,42 @@ char buf[3],*s;
 	kungetbuf = zalloc(kungetsz = 32);
 	kungetct = 0;
 	xbindtab = newhtable(67);
-	addhperm("\33\133C",makefunckey(z_forwardchar),xbindtab,NULL);
-	addhperm("\33\133D",makefunckey(z_backwardchar),xbindtab,NULL);
-	addhperm("\33\133A",makefunckey(z_uplineorhistory),xbindtab,NULL);
-	addhperm("\33\133B",makefunckey(z_downlineorhistory),xbindtab,NULL);
-	addhperm("\30*",makefunckey(z_expandword),xbindtab,NULL);
-	addhperm("\30g",makefunckey(z_listexpand),xbindtab,NULL);
-	addhperm("\30G",makefunckey(z_listexpand),xbindtab,NULL);
-	addhperm("\30\16",makefunckey(z_infernexthistory),xbindtab,NULL);
-	addhperm("\30\13",makefunckey(z_killbuffer),xbindtab,NULL);
-	addhperm("\30\6",makefunckey(z_vifindnextchar),xbindtab,NULL);
-	addhperm("\30\17",makefunckey(z_overwritemode),xbindtab,NULL);
-	addhperm("\30\25",makefunckey(z_undo),xbindtab,NULL);
-	addhperm("\30\26",makefunckey(z_vicmdmode),xbindtab,NULL);
-	addhperm("\30\12",makefunckey(z_vijoin),xbindtab,NULL);
-	addhperm("\30\2",makefunckey(z_vimatchbracket),xbindtab,NULL);
+	addhperm("\33\133C",makefunckey(z_forwardchar),xbindtab,(FFunc) 0);
+	addhperm("\33\133D",makefunckey(z_backwardchar),xbindtab,(FFunc) 0);
+	addhperm("\33\133A",makefunckey(z_uplineorhistory),xbindtab,(FFunc) 0);
+	addhperm("\33\133B",makefunckey(z_downlineorhistory),xbindtab,(FFunc) 0);
+	addhperm("\30*",makefunckey(z_expandword),xbindtab,(FFunc) 0);
+	addhperm("\30g",makefunckey(z_listexpand),xbindtab,(FFunc) 0);
+	addhperm("\30G",makefunckey(z_listexpand),xbindtab,(FFunc) 0);
+	addhperm("\30\16",makefunckey(z_infernexthistory),xbindtab,(FFunc) 0);
+	addhperm("\30\13",makefunckey(z_killbuffer),xbindtab,(FFunc) 0);
+	addhperm("\30\6",makefunckey(z_vifindnextchar),xbindtab,(FFunc) 0);
+	addhperm("\30\17",makefunckey(z_overwritemode),xbindtab,(FFunc) 0);
+	addhperm("\30\25",makefunckey(z_undo),xbindtab,(FFunc) 0);
+	addhperm("\30\26",makefunckey(z_vicmdmode),xbindtab,(FFunc) 0);
+	addhperm("\30\12",makefunckey(z_vijoin),xbindtab,(FFunc) 0);
+	addhperm("\30\2",makefunckey(z_vimatchbracket),xbindtab,(FFunc) 0);
 	addhperm("\30s",makefunckey(z_historyincrementalsearchforward),
-		xbindtab,NULL);
+		xbindtab,(FFunc) 0);
 	addhperm("\30r",makefunckey(z_historyincrementalsearchbackward),
-		xbindtab,NULL);
-	addhperm("\30u",makefunckey(z_undo),xbindtab,NULL);
+		xbindtab,(FFunc) 0);
+	addhperm("\30u",makefunckey(z_undo),xbindtab,(FFunc) 0);
 	addhperm("\30\30",makefunckey(z_exchangepointandmark),
-		xbindtab,NULL);
-	addhperm("run-help",mkanode(ztrdup("man"),1),aliastab,NULL);
-	addhperm("which-command",mkanode(ztrdup("whence"),1),aliastab,NULL);
+		xbindtab,(FFunc) 0);
+	addhperm("run-help",mkanode(ztrdup("man"),1),aliastab,(FFunc) 0);
+	addhperm("which-command",mkanode(ztrdup("whence"),1),aliastab,(FFunc) 0);
 	strcpy(buf,"\33q");
 	for (t0 = 128; t0 != 256; t0++)
 		if (emacsbind[t0] != z_undefinedkey) {
 			buf[1] = t0 & 0x7f;
-			addhnode(ztrdup(buf),makefunckey(emacsbind[t0]),xbindtab,NULL);
+			addhnode(ztrdup(buf),makefunckey(emacsbind[t0]),xbindtab,(FFunc) 0);
 		}
 	for (t0 = 0; t0 != 36; t0++) vibuf[t0] = NULL;
 	for (t0 = 0; t0 != 26; t0++) vimarkline[t0] = 0;
 	stackhist = stackcs = -1;
 	vichgbufsz = 0;
 	vichgbuf = NULL;
+	usernamescached=0;
 }
 
 char *getkeystring(s,len) /**/
@@ -422,7 +439,7 @@ char *s;int *len;
 {
 static char buf[512];
 char *t = buf;
-int x,first = 1,metanext = 0;
+int x,metanext = 0;
 
 	for (;*s;s++)
 		{
@@ -468,7 +485,6 @@ int x,first = 1,metanext = 0;
 			}
 		if (t > buf+500)
 			break;
-		first = 0;
 		}
 	*t = '\0';
 	*len = t-buf;
@@ -520,7 +536,7 @@ char *str;Key k;
 int bin_bindkey(name,argv,ops,junc) /**/
 char *name;char **argv;char *ops;int junc;
 {
-int t0,len;
+int t0 = 0,len;
 char *s;
 int func,*tab;
 
