@@ -68,21 +68,6 @@ void backwarddeletechar() /**/
 	backdel(mult);
 }
 
-void videletechar() /**/
-{
-	if (mult < 0) { mult = -mult; vibackwarddeletechar(); return; }
-	if (c == 4 && !ll) {
-		eofsent = 1;
-		return;
-	}
-	if (!(cs+mult > ll || line[cs] == '\n')) {
-		cs += mult;
-		backkill(mult,0);
-		if (cs && (cs == ll || line[cs] == '\n')) cs--;
-	} else
-		feep();
-}
-
 void vibackwarddeletechar() /**/
 {
 	if (mult < 0) { mult = -mult; videletechar(); return; }
@@ -153,26 +138,35 @@ int cc;
 
 void transposechars() /**/
 {
-int cc;
+int cc, ct;
 int neg = mult < 0;
 
 	if (neg) mult = -mult;
 	while (mult--) {
-		if (cs == 0 || line[cs-1] == '\n') {
-			if (ll == cs || line[cs] == '\n' || line[cs+1] == '\n') {
+		if (!(ct = cs) || line[cs-1] == '\n') {
+			if (ll == cs || line[cs] == '\n') {
 				feep();
 				return;
 			}
-			cs++;
+			if (!neg) cs++;
+		   ct++;
 		}
-		if (!neg) {
-			if (cs != ll && line[cs] != '\n') cs++;
+		if (neg) {
+			if (cs && line[cs-1] != '\n') {
+			  cs--;
+			  if (ct > 1 && line[ct-2] != '\n') ct--;
+			}
 		} else {
-			if (cs != 0 && line[cs-1] != '\n') cs--;
+			if (cs != ll && line[cs] != '\n') cs++;
 		}
-		cc = line[cs-2];
-		line[cs-2] = line[cs-1];
-		line[cs-1] = cc;
+		if (ct == ll || line[ct] == '\n') ct--;
+		if (ct < 1 || line[ct-1] == '\n') {
+		   feep();
+			return;
+		}
+		cc = line[ct-1];
+		line[ct-1] = line[ct];
+		line[ct] = cc;
 	}
 }
 
@@ -349,15 +343,16 @@ struct sgttyb sob;
 
 void digitargument() /**/
 {
-	if (!(lastcmd & ZLE_ARG))
+	int sign = (mult < 0 || (lastcmd & ZLE_NEGARG)) ? -1 : 1;
+	if ((lastcmd & (ZLE_ARG|ZLE_NEGARG)) != ZLE_ARG)
 		mult = 0;
-	mult = mult*10+(c&0xf);
-	if (lastcmd & ZLE_NEGARG) mult = -mult;
+	mult = mult*10 + sign*(c&0xf);
 }
 
 void negargument() /**/
 {
 	if (lastcmd & ZLE_ARG) feep();
+	mult = -1;
 }
 
 void universalargument() /**/
@@ -524,7 +519,7 @@ int len,ch,t0;
 							ambig = xx;
 						}
 				permalloc();
-				if (!full(ll))
+				if (empty(ll))
 					feep();
 				else if (!nextnode(firstnode(ll)))
 					{
@@ -644,28 +639,43 @@ int d;
 
 /* get a prompt string */
 
-char *putprompt(fm,lenp) /**/
-char *fm;int *lenp;
+char *putprompt(fm,lenp,isspell) /**/
+char *fm;int *lenp;int isspell;
 {
 char *ss,*bl0;
-static char buf1[256],buf2[256],*buf;
+static char buf0[256],buf1[256],buf2[256],*buf;
 char buf3[MAXPATHLEN];
-int t0,bracepos = 0;
-struct tm *tm = NULL;
+int t0,bracepos = 0,arg;
+struct tm *tm;
 time_t timet;
 
 	lensb = 0; countp = 1;
 	if (!fm) { *lenp = 0; return ""; }
-	/* kludge alert! */
-	buf = (buf == buf1) ? buf2 : buf1;
+	/* KLUDGE ALERT!  What we have here are three buffers:
+	 *  buf1 and buf2 alternate between PS1 and PS2, though which is
+	 *   which is indeterminate depending on spellchecking, "select",
+	 *   etc. -- those operations also share these two buffers.
+	 *  buf0 is used for any prompting that manages to happen while
+	 *   zleread() is in progress (signal traps, etc.), because
+	 *   zleread() re-uses the pointers returned to buf1 and buf2
+	 *   and will be confused if either of those is overwritten.
+	 */
+	buf = zleactive ? buf0 : ((buf == buf1) ? buf2 : buf1);
 	bp = bl0 = buf;
 	if (!columns) columns = 80;
 	clearerr(stdin);
 	for(;*fm;fm++) {
 		if (bp-buf >= 220)
 			break;
+		arg = 0;
 		if (*fm == '%')
-			switch (*++fm) {
+			{
+			if (idigit(*++fm))
+				{
+				arg = zstrtol(fm, &fm, 10);
+				}
+			switch (*fm)
+				{
 				case '~':
 					t0 = finddir(pwd);
 					if (t0 != -1) {
@@ -674,26 +684,18 @@ time_t timet;
 						stradd(pwd+strlen(userdirs[t0]));
 						break;
 					}
-					if (!strncmp(pwd,home,t0 = strlen(home)) && t0 > 1) {
-						*bp++ = '~';
-						stradd(pwd+t0);
-						break;
-					}
 				case 'd': case '/': stradd(pwd); break;
 				case 'c': case '.':
 					t0 = finddir(pwd);
 					if (t0 != -1) {
 						sprintf(buf3,"~%s%s",usernames[t0],
 							pwd+strlen(userdirs[t0]));
-					} else if (!strncmp(pwd,home,t0 = strlen(home)) && t0 > 1) {
-						sprintf(buf3,"~%s",pwd+t0);
 					} else {
 						strcpy(buf3,pwd);
 					}
-					t0 = 1;
-					if (idigit(fm[1])) { t0 = fm[1]-'0'; fm++; }
+					if (!arg) arg++;
 					for (ss = buf3+strlen(buf3); ss > buf3; ss--)
-						if (*ss == '/' && !--t0) {
+						if (*ss == '/' && !--arg) {
 							ss++;
 							break;
 						}
@@ -702,10 +704,9 @@ time_t timet;
 					break;
 				case 'C':
 					strcpy(buf3,pwd);
-					t0 = 1;
-					if (idigit(fm[1])) { t0 = fm[1]-'0'; fm++; }
+					if (!arg) arg++;
 					for (ss = buf3+strlen(buf3); ss > buf3; ss--)
-						if (*ss == '/' && !--t0) {
+						if (*ss == '/' && !--arg) {
 							ss++;
 							break;
 						}
@@ -718,12 +719,9 @@ time_t timet;
 					break;
 				case 'M': stradd(hostnam); break;
 				case 'm':
-					if (idigit(fm[1]))
-						t0 = (*++fm)-'0';
-					else
-						t0 = 1;
+					if (!arg) arg++;
 					for (ss = hostnam; *ss; ss++)
-						if (*ss == '.' && !--t0)
+						if (*ss == '.' && !--arg)
 							break;
 					t0 = *ss;
 					*ss = '\0';
@@ -803,9 +801,33 @@ time_t timet;
 					break;
 				case '%': *bp++ = '%'; break;
 				case '#': *bp++ = (geteuid()) ? '%' : '#'; break;
-				case 'r': stradd(rstring); break;
-				case 'R': stradd(Rstring); break;
-				default: *bp++ = '%'; *bp++ = *fm; break;
+				case 'v':
+					if (!arg) arg++;
+					/* The number 35 here comes from 256-220-1, where 256 is
+					 * sizeof(buf), 220 is from the overflow test made at the
+					 * top of the loop, and 1 is for the \0 byte at the end.
+					 */
+					if (arrlen(psvar) >= arg && strlen(psvar[arg-1]) < 35)
+						stradd(psvar[arg-1]);
+					else
+						stradd("");
+					break;
+				case 'r':
+					if (isspell)
+						{
+						stradd(rstring);
+						break;
+						}
+				case 'R':
+					if (isspell)
+						{
+						stradd(Rstring);
+						break;
+						}
+				default:
+					*bp++ = '%'; *bp++ = *fm;
+					break;
+				}
 			}
 		else if (*fm == '!') {
 			sprintf(bp,"%d",curhist);
