@@ -131,7 +131,7 @@ int x,y,ocs = cs;
 void vimatchbracket() /**/
 {
 int ocs = cs,dir,ct;
-char oth,me;
+unsigned char oth,me;
 
 otog:
 	if (cs == ll)
@@ -318,7 +318,6 @@ int ch;
 		if (curhist != vimarkline[ch]) return;
 	}
 	cs = vimarkcs[ch];
-	if (cs > ll) ch = ll;
 }
 
 void vigotomarkline() /**/
