@@ -108,7 +108,7 @@ int t0,hwinh = winh/2;
 void refresh() /**/
 {
 unsigned char *s,*t,*sen,*scs = line+cs; char **qbuf;
-int ln = 0,nvcs,nvln = -1,t0;
+int ln = 0,nvcs = 0,nvln = -1,t0;
 
 	cost = 0;
 	if (resetneeded)
@@ -326,7 +326,7 @@ int ccs = 0;
 				moveto(ln,ccs);
 				fwrite(nl,strlen(nl),1,stdout);
 				cost += strlen(nl);
-				ccs = (vcs += strlen(nl));
+				vcs += strlen(nl);
 				}
 			*truncptr = 0;
 			return;
@@ -459,7 +459,7 @@ int ct;
 /* do a multright if it's cheaper or if we're walking over the prompt.  */
 
 	if (tccan(TCMULTRIGHT) &&
-			(ct > tclen[TCMULTRIGHT] || vln == 0 && vcs < pptlen))
+			(ct > tclen[TCMULTRIGHT] || (vln == 0 && vcs < pptlen)))
 		tcoutarg(TCMULTRIGHT,ct);
 
 /* if we're walking over the prompt and we can do a bunch of cursor rights,
@@ -534,7 +534,7 @@ void trashzle() /**/
 void singlerefresh() /**/
 {
 char *vbuf,*vp,**qbuf,*op;
-int t0,vsiz,nvcs;
+int t0,vsiz,nvcs = 0;
 
 	for (vsiz = 1+pptlen, t0 = 0; t0 != ll; t0++,vsiz++)
 		if (line[t0] == '\t')
