@@ -20,16 +20,20 @@
 
 #define ZLE
 #include "zsh.h"
-#ifdef __hpux
-#include <ndir.h>
-#else
-#ifdef SYSV
-#define direct dirent
-#else
-#include <sys/dir.h>
-#endif
-#endif
 #include	<pwd.h>
+#ifdef HAS_NIS_PASSWD
+#include	<rpc/rpc.h>
+#include	<rpcsvc/ypclnt.h>
+#include	<rpcsvc/yp_prot.h>
+
+#define PASSWD_FILE	"/etc/passwd"
+#define PASSWD_MAP	"passwd.byname"
+
+typedef struct {
+	int len;
+	char *s;
+	} dopestring;
+#endif
 
 static int we,wb,usemenu,useglob;
 
@@ -131,8 +135,6 @@ void listexpand() /**/
 
 void reversemenucomplete() /**/
 {
-char *s;
-
 	if (!menucmp)
 		menucompleteword();	/* better than just feep'ing, pem */
 	if (!menucmp) return;
@@ -142,7 +144,7 @@ char *s;
 		menunode = lastnode(menulist);
 	else
 		menunode = prevnode(menunode);
-	inststr(s = menunode->dat);
+	inststr(menunode->dat);
 	menue = cs;
 }
 
@@ -212,6 +214,8 @@ char *s;
 		}
 		if (lst == COMP_SPELL) {
 			char	**x = &s;
+			char *q = s;
+			for(; *q; q++) if (INULL(*q)) *q = Nularg;
 			untokenize(s);
 			cs = wb;
 			foredel(we-wb);
@@ -260,7 +264,7 @@ char *s;
 	inststr(s = menunode->dat);
 	if (isset(LASTMENU)) {
 		if (lastmenu) free(lastmenu);
-		lastmenu = ztrdup(line);
+		lastmenu = ztrdup(UTOSCP(line));
 		lastmenupos = cs;
 	}
 	menue = cs;
@@ -269,7 +273,7 @@ char *s;
 char *get_comp_string() /**/
 {
 int t0;
-unsigned char *s,*linptr;
+unsigned char *s = NULL,*linptr;
 
 	linptr = line;
 start:
@@ -279,7 +283,7 @@ start:
 	zleparse = 1;
 	lexsave();
 	hungets(" "); /* KLUDGE! */
-	hungets(linptr);
+	hungets(UTOSCP(linptr));
 	strinbeg();
 	pushheap();
 	do {
@@ -298,7 +302,7 @@ start:
 		s = (unsigned char *)ztrdup(tokstr);
 	} else if (t0 == ENVSTRING) {
 		for (s = (unsigned char *)tokstr; *s && *s != (unsigned char)'='; s++, wb++);
-		if (*s) { s++; wb++; t0 = STRING; s = (unsigned char *)ztrdup(s); }
+		if (*s) { s++; wb++; t0 = STRING; s = STOUCP(ztrdup(UTOSCP(s))); }
 		lincmd = 1;
 	}
 	hflush();
@@ -329,7 +333,7 @@ char *ss;
 	postfork(vl,1);
 	if (errflag)
 		goto end;
-	if (!full(vl) || !*(char *) peekfirst(vl)) {
+	if (empty(vl) || !*(char *) peekfirst(vl)) {
 		feep();
 		goto end;
 	}
@@ -451,23 +455,99 @@ char *s;char *t;
 	if (strpfx(pat,s) && ((Cmdnam) t)->type != DISABLED) addmatch(s);
 }
 
+#ifdef HAS_NIS_PASSWD
+static int match_username(status, key, keylen, val, vallen, data)
+int status;
+char *key, *val;
+int keylen, vallen;
+dopestring *data;
+{
+	if (errflag || status != YP_TRUE) return 1;
+
+	if (vallen > keylen && val[keylen] == ':')
+		{
+		val[keylen] = '\0';
+		/* Can't call getpwnam() here; it breaks yp_all() */
+		if(strncmp(val,data->s,data->len) == 0) addmatch(val);
+		}
+	return 0;
+}
+#endif
+
 void maketildelist(s) /**/
 char	*s;
 {
+#ifdef HAS_NIS_PASSWD
+	char domain[YPMAXDOMAIN];
+	struct ypall_callback cb;
+	dopestring data;
+	Lknode n;
+	FILE *pwf;
+	char buf[BUFSIZ], *p;
+	int skipping;
+	
+	data.s = ++s;
+	data.len = strlen(s);
+	if(*s == '+' || *s == '-')
+		{
+		/* We don't want to match any NIS inclusions/exclusions */
+		*s = '\0';
+		return;
+		}
+	/* Get potential matches from NIS and cull those without local accounts */
+	if (getdomainname(domain, YPMAXDOMAIN) == 0)
+		{
+		cb.foreach = match_username;
+		cb.data = (char *) &data;
+		yp_all(domain, PASSWD_MAP, &cb);
+		for(n = firstnode(matches); n; incnode(n))
+			if(getpwnam(getdata(n)) == NULL) uremnode(matches,n);
+		}
+	/* Don't forget the non-NIS matches from the flat passwd file */
+	if ((pwf = fopen(PASSWD_FILE, "r")) != NULL)
+		{
+		skipping = 0;
+		while (fgets(buf, BUFSIZ, pwf) != NULL)
+			{
+			if (strchr(buf, '\n') != NULL)
+				{
+				if (!skipping)
+					{
+					if (strncmp(buf,data.s,data.len) == 0 &&
+							(p = strchr(buf, ':')) != NULL)
+						{
+						*p = '\0';
+						addmatch(buf);
+						}
+					}
+				else skipping = 0;
+				}
+			else skipping = 1;
+			}
+		fclose(pwf);
+		}
+#else
 	struct passwd	*pwd;
-	int		len;
+	int		len,i;
+
+	if (!usernamescached)
+		{
+		setpwent();
+		while((pwd=getpwent()) != NULL && !errflag)
+			adduserdir(pwd->pw_name,pwd->pw_dir);
+		endpwent();
+		usernamescached=1;
+		}
 
 	s++;
 	len = strlen(s);
-	if (len < 1) {
-		addmatch(s);
-		*s = 0;
-		return;
-	}
-	while ((pwd = getpwent()) != NULL && !errflag)
-		if (strncmp(pwd->pw_name, s, len) == 0)
-			addmatch(pwd->pw_name);
-	endpwent();
+
+	for(i=0;i<userdirsz;i++)
+		{
+		if(usernames[i] && userdirs[i] &&
+			strncmp(usernames[i],s,len)==0) addmatch(usernames[i]);
+		}
+#endif
 	*s = 0;
 }
 
@@ -538,9 +618,10 @@ char *pathprefix;
 void docompletion(s,lst,incmd) /**/
 char *s;int lst;int incmd;
 {
-char *tokorigs;
+char *tokorigs = NULL;
 char *origs;
 Compctl cc;
+char *u;
 char *pfx = s;
 
 	slashflag = SLASH_MAYBE;
@@ -568,7 +649,7 @@ char *pfx = s;
 	if (cc->mask & (CC_FILES|CC_COMMPATH)) {
 		/* only do "globbed" completion if regular completion fails.
 		   pem, 7Oct91 */
-		if ((!full(matches) || errflag) && useglob) {
+		if ((empty(matches) || errflag) && useglob) {
 		    gen_matches_glob(tokorigs,incmd);
 		    /*
 		     * gen_matches_glob changes the insert line to be correct up
@@ -581,7 +662,25 @@ char *pfx = s;
 	if ((cc->mask & CC_HOSTS) && !haspath) {
 		char **x;
 		for (x = hosts; *x; x++) addcmdmatch(*x,NULL);
+
+/* we now try to do expansion if there is an @ symbol in the string */
+
+		haspath = 0;
+		for (u = s+strlen(s); u >= s; u--)
+			if (*u == '@' ) break;
+		if (u >= s) {
+			typechar = *u;
+			*u++ = '\0';
+			haspath = 1;
+		} else u = s;
+		pat = u;
+		if (typechar == '@' && haspath ) {
+			char **x;
+			slashflag = SLASH_NO;
+			for (x = hosts; *x; x++) addcmdmatch(*x,NULL);
+		}
 	}
+	pat = s;
 	if ((cc->mask & CC_OPTIONS) && !haspath) {
 		struct option *o;
 		for (o = optns; o->name; o++) addcmdmatch(o->name,NULL);
@@ -597,7 +696,7 @@ char *pfx = s;
 		if (usr) while (*usr) addcmdmatch(*usr++,NULL);
 	}
 	if (lst != COMP_LIST_COMPLETE) do_fignore(origs);
-	if (!full(matches) || errflag) {
+	if (empty(matches) || errflag) {
 		feep();
 	} else if (lst == COMP_LIST_COMPLETE) {
 		listmatches(matches,
@@ -676,7 +775,7 @@ struct direct *de;
 		 if (hasp) {
 			/* Find the longest common prefix string
 			 * after globbing the input. All expansions
-			 * ~foo/bar/* will turn into something like
+			 * '~foo/bar/*' will turn into something like
 			 * /tmp_mnt/hosts/somehost/home/foo/...
 			 * We will remove this common prefix from the matches.
 			 * ash, 7 May '91
@@ -702,7 +801,7 @@ struct direct *de;
 		prefork(matches);
 		if (!errflag) postfork(matches,1);
 		opts[NONOMATCH] = nonomatch;
-		if (errflag || !full(matches) || !nextnode(firstnode(matches))) {
+		if (errflag || empty(matches) || !nextnode(firstnode(matches))) {
 			/* if there were no matches (or only one)
 				add a trailing * and try again */
 			s = dyncat(s,"x");
@@ -822,7 +921,6 @@ char *origstr;
 	if (full(matches) && nextnode(firstnode(matches))) {
 		Lknode z,zn;
 
-		z = firstnode(matches);
 		ambig = 1000;
 		for (z = firstnode(matches); z; z = zn) {
 			char *q = getdata(z);
@@ -969,7 +1067,7 @@ char *s;
 	if (isset(LASTMENU)) {
 		if (lastmenu)
 			free(lastmenu);
-		lastmenu = ztrdup(line);
+		lastmenu = ztrdup(UTOSCP(line));
 		lastmenupos = cs;
 	}
 }
@@ -1025,7 +1123,8 @@ char **arr,**ap;
 		singsub(&apps);
 		longest++;
 		}
-	qsort(arr,ct,sizeof(char *),forstrcmp);
+	qsort((vptr)arr,ct,sizeof(char *),
+			(int (*) DCLPROTO((const void *, const void *)))forstrcmp);
 	fct = (columns-1)/(longest+2);
 	if (fct == 0)
 		fct = 1;
@@ -1159,7 +1258,7 @@ int t0,oldcs,oldll;
 	oldll = ll;
 	zleparse = 1;
 	lexsave();
-	hungets(line);
+	hungets(UTOSCP(line));
 	strinbeg();
 	pushheap();
 	ll = cs = 0;
@@ -1199,13 +1298,13 @@ static int cmdwb,cmdwe;
 
 char *getcurcmd() /**/
 {
-int lincmd = incmdpos;
+int lincmd;
 char *s = NULL;
 
 	zleparse = 1;
 	lexsave();
 	hungets(" "); /* KLUDGE! */
-	hungets(line);
+	hungets(UTOSCP(line));
 	strinbeg();
 	pushheap();
 	do {
@@ -1217,7 +1316,6 @@ char *s = NULL;
 			s = ztrdup(tokstr);
 			cmdwb = ll-wordbeg; cmdwe = ll+1-inbufct;
 		}
-		lincmd = incmdpos;
 	} while (tok != ENDINPUT && zleparse);
 	hflush();
 	popheap();
