@@ -138,25 +138,34 @@ void vibackwardkillword() /**/
 {
 int x = cs;
 
-	if (mult < 0) { feep(); return; }
-	while (mult--) {
-		while (x > viinsbegin && !iword(line[x-1])) x--;
-		while (x > viinsbegin && iword(line[x-1])) x--;
-	}
-	backkill(cs-x,1);
+        if (mult < 0) { feep(); return; }
+	/* this taken from "vibackwardword" */
+        while (mult--) {
+                while ((x > viinsbegin) && iblank(line[x-1])) x--;
+                if (iident(line[x-1])) while ((x > viinsbegin) && iident(line[x-1])) x--;
+                else while ((x > viinsbegin) && !iident(line[x-1]) && !iblank(line[x-1])) x--;
+        }
+        /* 
+        while (mult--) {
+                while ( (x > viinsbegin) && (iwordsep(line[x-1]))) x--;
+                while ( (x > viinsbegin) && (!iwordsep(line[x-1]))) x--;
+        }
+        */ 
+        backkill(cs-x,1);
 }
 
 void backwardkillword() /**/
 {
 int x = cs;
 
-	if (mult < 0) { mult = -mult; killword(); return; }
-	while (mult--) {
-		while (x && !iword(line[x-1])) x--;
-		while (x && iword(line[x-1])) x--;
-	}
-	backkill(cs-x,1);
-}
+        if (mult < 0) { mult = -mult; killword(); return; }
+        while (mult--) {                       
+                while (x && !iword(line[x-1])) x--;
+                while (x && iword(line[x-1])) x--;
+        }                                      
+        backkill(cs-x,1);                      
+}                                              
+
 
 void upcaseword() /**/
 {
@@ -262,9 +271,9 @@ int neg = mult < 0, ocs = cs;
 		}
 		for (p1 = p2; p1 && iword(line[p1-1]); p1--);
 		pp = temp = halloc(p4-p1+1);
-		struncpy(&pp,line+p3,p4-p3);
-		struncpy(&pp,line+p2,p3-p2);
-		struncpy(&pp,line+p1,p2-p1);
+		struncpy(&pp,UTOSCP(line+p3),p4-p3);
+		struncpy(&pp,UTOSCP(line+p2),p3-p2);
+		struncpy(&pp,UTOSCP(line+p1),p2-p1);
 		strncpy((char *) line+p1,temp,p4-p1);
 		cs = p4;
 	}
