@@ -29,6 +29,10 @@
 #include <strings.h>
 #endif
 
+#ifdef HAS_MEMORY
+#include <memory.h>
+#endif
+
 #ifdef HAS_LOCALE
 #include <locale.h>
 #endif
@@ -38,15 +42,16 @@
 #endif
 
 #ifdef SYSV
+#ifndef SYSVR4
 #include <sys/bsdtypes.h>
 #define _POSIX_SOURCE
 #include <sys/limits.h>
 #include <sys/sioctl.h>
 #define MAXPATHLEN PATH_MAX
 #define lstat stat
+#endif
 extern int gethostname();
 #define sigmask(m) m
-#include <sys/dirent.h>
 #else
 #include <sys/types.h>		/* this is the key to the whole thing */
 #endif
@@ -56,16 +61,29 @@ extern int gethostname();
 #include <sys/wait.h>
 #define _BSD
 #else
+#ifdef HAS_WAIT
+#include <wait.h>
+#else
 #include <sys/wait.h>
 #endif
+#endif
 
+#ifdef HAS_TIME
+#include <time.h>
+#else
 #include <sys/time.h>
+#endif
 
-#ifndef SYSV
+#ifdef SYSV
+#include <fcntl.h>
+#else
+#include <sys/file.h>
+#endif
+
+#if !defined(SYSV) || defined(SYSVR4)
 #include <sys/resource.h>
 #endif
 
-#include <sys/file.h>
 #include <signal.h>
 
 #ifdef TERMIO
@@ -82,13 +100,21 @@ extern int gethostname();
 #endif
 #endif
 
-#ifdef SYSV
+#if defined(SYSV) && !defined(SYSVR4)
+#define readlink(s,t,z)	(-1)
 #undef TIOCGWINSZ
 #endif
 
 #include <sys/param.h>
 #include <sys/times.h>
 
+#ifdef HAS_DIRENT
+#include <dirent.h>
+#define direct dirent
+#else
+#include <sys/dir.h>
+#endif
+
 #ifdef SYSV
 #undef _POSIX_SOURCE
 #endif
@@ -97,31 +123,47 @@ extern int gethostname();
 #include <sys/bsdtty.h>
 #endif
 
-#ifndef sun
+#if !defined(sun) && !defined(SYSVR4)
 #include <sys/ioctl.h>
 #else
 #include <sys/filio.h>
 #endif
 
-#define VERSIONSTR "zsh v2.2.0"
-
-#if 0 /* __STDC__ */
-#include <unistd.h>
-#include <fcntl.h>
-#include <stat.h>
+#ifdef __STDC__
 #define DCLPROTO(X) X
+#ifdef HAS_UNISTD
+#include <unistd.h>
+#endif
+#include <fcntl.h>
+#include <sys/stat.h>
 #undef NULL
 #define NULL ((void *)0)
-#else /* not __STDC__ */
-#include <sys/stat.h>
+#else /* K&R C */
 #define DCLPROTO(X) ()
+#define const
+#include <sys/stat.h>
 #ifndef NULL
 #define NULL 0
 #endif
-#endif /* __STDC__ */
+#endif
 
-#ifdef QDEBUG
-#define DCLPROTO(X) X
+#ifdef HAS_UTMPX
+#include <utmpx.h>
+# define utmp utmpx
+# define ut_time ut_xtime
+# undef UTMP_FILE
+# define UTMP_FILE UTMPX_FILE
+# undef WTMP_FILE
+# define WTMP_FILE WTMPX_FILE
+#else
+#include <utmp.h>
+#endif
+
+#if !defined(UTMP_FILE) && defined(_PATH_UTMP)
+#define UTMP_FILE _PATH_UTMP
+#endif
+#if !defined(WTMP_FILE) && defined(_PATH_WTMP)
+#define WTMP_FILE _PATH_WTMP
 #endif
 
 #define DEFWORDCHARS "*?_-.[]~=/&;!#$%^(){}<>"
@@ -142,6 +184,11 @@ extern int gethostname();
 #define killpg(pgrp,sig) kill(-(pgrp),sig)
 #endif
 
+/* SYSV or POSIX compatible BSD */
+#if defined(SYSV) || ( defined(BSD) && BSD >= 199103 )
+#define getpgrp(X)  getpgrp()
+#endif
+
 #ifndef F_OK
 #define F_OK 00
 #define R_OK 04
@@ -187,7 +234,40 @@ vptr (*ncalloc)DCLPROTO((int));
 #define Tilde			((char) 0x95)
 #define Qtick			((char) 0x96)
 #define Comma			((char) 0x97)
-#define Nularg			((char) 0x98)
+#define Snull			((char) 0x98)
+#define Dnull			((char) 0x99)
+#define Bnull			((char) 0x9a)
+#define Nularg			((char) 0x9b)
+
+#define INULL(x)		(((x) & 0x98) == 0x98)
+
+/* Character tokens are sometimes casted to (unsigned char)'s. Unfortunately,
+   SVR4's deceiving compiler botches non-terminal, same size, signed to unsigned
+	promotions; i.e. (int) (unsigned char) ((char) -1) evaluates to -1,
+	not 255 as it should!
+
+	We circumvent the troubles of such shameful delinquency by casting to a
+	larger unsigned type then back down to unsigned char.
+		-- Marc Boucher <marc@cam.org>
+	<daniel@ug.eds.com> 25-Jan-1993  Dec Alpha OSF compilers have the same property
+*/
+
+#if (defined(SYSVR4) || defined (__osf__)) && !defined(__GNUC__)
+#define STOUC(X)	((unsigned char)(unsigned short)(X))
+#else
+#define STOUC(X)	((unsigned char)(X))
+#endif
+
+/*
+ * The source was full of implicit casts between signed and unsigned
+ * character pointers.  To get a clean compile, I've made these casts
+ * explicit, but the potential for error is still there.  If your machine
+ * needs special treatment, just define your own macros here.
+ * 	--jim <jmattson@ucsd.edu>
+ */
+ 
+#define STOUCP(X)	((unsigned char *)(X))
+#define UTOSCP(X)	((char *)(X))
 
 /* chars that need to be quoted if meant literally */
 
@@ -293,7 +373,15 @@ int redirtab[TRINANG-OUTANG+1] = {
 	HERESTR,
 };
 #else
-int redirtab[TRINANG-OUTANG+1];
+extern int redirtab[TRINANG-OUTANG+1];
+#endif
+
+#ifdef GLOBALS
+char nulstring[] = { Nularg, '\0' };
+int nulstrlen = sizeof(nulstring) - 1;
+#else
+extern char nulstring[];
+extern int nulstrlen;
 #endif
 
 #define IS_READFD(X) ((X)>=READ && (X)<=MERGE)
@@ -305,7 +393,7 @@ int redirtab[TRINANG-OUTANG+1];
 #define FD_COPROC -2
 #define FD_CLOSE  -3
 
-EXTERN char **environ;
+extern char **environ;
 
 typedef struct hashtab *Hashtab;
 typedef struct hashnode *Hashnode;
@@ -330,7 +418,6 @@ typedef struct redir *Redir;
 typedef struct complist *Complist;
 typedef struct heap *Heap;
 typedef struct histent *Histent;
-typedef struct hp *Hp;
 typedef struct compctl *Compctl;
 typedef void (*FFunc)DCLPROTO((vptr));
 typedef vptr (*VFunc)DCLPROTO((vptr));
@@ -351,6 +438,7 @@ struct lklist {
 
 #define addnode(X,Y) insnode(X,(X)->last,Y)
 #define full(X) ((X)->first != NULL)
+#define empty(X) ((X)->first == NULL)
 #define firstnode(X) ((X)->first)
 #define getaddrdata(X) (&((X)->dat))
 #define getdata(X) ((X)->dat)
@@ -756,8 +844,11 @@ struct histent {
 	char *lit;			/* literal history line */
 	time_t stim;		/* command started time (datestamp) */
 	time_t ftim;		/* command finished time */
+	int flags;		/* Misc flags */
 	};
 
+#define HIST_OLD	0x00000001	/* Command is already written to disk*/
+
 /* completion control */
 
 struct compctl {
@@ -868,90 +959,94 @@ struct option {
 #define NOHISTBEEP '\32'
 #define PUSHDIGNOREDUPS '\33'
 #define AUTOREMOVESLASH '\34'
+#define EXTENDEDHISTORY '\35'
+#define APPENDHISTORY '\36'
 
 #ifndef GLOBALS
 extern struct option optns[];
 #else
 struct option optns[] = {
-	"correct",CORRECT,
-	"noclobber",NOCLOBBER,
-	"nobadpattern",NOBADPATTERN,
-	"nonomatch",NONOMATCH,
-	"globdots",GLOBDOTS,
-	"notify",NOTIFY,
-	"bgnice",BGNICE,
-	"ignoreeof",IGNOREEOF,
-	"markdirs",MARKDIRS,
-	"autolist",AUTOLIST,
-	"nobeep",NOBEEP,
-	"printexitvalue",PRINTEXITVALUE,
-	"pushdtohome",PUSHDTOHOME,
-	"pushdsilent",PUSHDSILENT,
-	"noglob",NOGLOBOPT,
-	"nullglob",NULLGLOB,
-	"rmstarsilent",RMSTARSILENT,
-	"ignorebraces",IGNOREBRACES,
-	"braceccl",BRACECCL,
-	"autocd",AUTOCD,
-	"nobanghist",NOBANGHIST,
-	"sunkeyboardhack",SUNKEYBOARDHACK,
-	"singlelinezle",SINGLELINEZLE,
-	"autopushd",AUTOPUSHD,
-	"correctall",CORRECTALL,
-	"rcexpandparam",RCEXPANDPARAM,
-	"pathdirs",PATHDIRS,
-	"longlistjobs",LONGLISTJOBS,
-	"recexact",RECEXACT,
-	"cdablevars",CDABLEVARS,
-	"mailwarning",MAILWARNING,
-	"nopromptcr",NOPROMPTCR,
-	"autoresume",AUTORESUME,
-	"listtypes",LISTTYPES,
-	"menucomplete",MENUCOMPLETE,
-	"zle",USEZLE,
-	"allexport",ALLEXPORT,
-	"errexit",ERREXIT,
-	"norcs",NORCS,
-	"histignorespace",HISTIGNORESPACE,
-	"histignoredups",HISTIGNOREDUPS,
-	"interactive",INTERACTIVE,
-	"histlit",HISTLIT,
-	"interactivecomments",INTERACTIVECOMMENTS,
-	"login",LOGINSHELL,
-	"monitor",MONITOR,
-	"noexec",NOEXEC,
-	"shinstdin",SHINSTDIN,
-	"nounset",NOUNSET,
-	"verbose",VERBOSE,
-	"chaselinks",CHASELINKS,
-	"xtrace",XTRACE,
-	"shwordsplit",SHWORDSPLIT,
-	"menucompletebeep",MENUCOMPLETEBEEP,
-	"histnostore",HISTNOSTORE,
-	"extendedglob",EXTENDEDGLOB,
-	"globcomplete",GLOBCOMPLETE,
-	"cshjunkiequotes",CSHJUNKIEQUOTES,
-	"pushdminus",PUSHDMINUS,
-	"cshjunkieloops",CSHJUNKIELOOPS,
-	"rcquotes",RCQUOTES,
-	"noshortloops",NOSHORTLOOPS,
-	"lastmenu",LASTMENU,
-	"automenu",AUTOMENU,
-	"histverify",HISTVERIFY,
-	"nolistbeep",NOLISTBEEP,
-	"nohup",NOHUP,
-	"noequals",NOEQUALS,
-	"kshoptionprint",KSHOPTIONPRINT,
-	"cshnullglob",CSHNULLGLOB,
-	"hashcmds",HASHCMDS,
-	"hashdirs",HASHDIRS,
-	"numericglobsort",NUMERICGLOBSORT,
-	"hashlistall",HASHLISTALL,
-	"overstrike",OVERSTRIKE,
-	"nohistbeep",NOHISTBEEP,
-	"pushdignoredups",PUSHDIGNOREDUPS,
-	"autoremoveslash",AUTOREMOVESLASH,
-	NULL,0
+	{"correct",CORRECT},
+	{"noclobber",NOCLOBBER},
+	{"nobadpattern",NOBADPATTERN},
+	{"nonomatch",NONOMATCH},
+	{"globdots",GLOBDOTS},
+	{"notify",NOTIFY},
+	{"bgnice",BGNICE},
+	{"ignoreeof",IGNOREEOF},
+	{"markdirs",MARKDIRS},
+	{"autolist",AUTOLIST},
+	{"nobeep",NOBEEP},
+	{"printexitvalue",PRINTEXITVALUE},
+	{"pushdtohome",PUSHDTOHOME},
+	{"pushdsilent",PUSHDSILENT},
+	{"noglob",NOGLOBOPT},
+	{"nullglob",NULLGLOB},
+	{"rmstarsilent",RMSTARSILENT},
+	{"ignorebraces",IGNOREBRACES},
+	{"braceccl",BRACECCL},
+	{"autocd",AUTOCD},
+	{"nobanghist",NOBANGHIST},
+	{"sunkeyboardhack",SUNKEYBOARDHACK},
+	{"singlelinezle",SINGLELINEZLE},
+	{"autopushd",AUTOPUSHD},
+	{"correctall",CORRECTALL},
+	{"rcexpandparam",RCEXPANDPARAM},
+	{"pathdirs",PATHDIRS},
+	{"longlistjobs",LONGLISTJOBS},
+	{"recexact",RECEXACT},
+	{"cdablevars",CDABLEVARS},
+	{"mailwarning",MAILWARNING},
+	{"nopromptcr",NOPROMPTCR},
+	{"autoresume",AUTORESUME},
+	{"listtypes",LISTTYPES},
+	{"menucomplete",MENUCOMPLETE},
+	{"zle",USEZLE},
+	{"allexport",ALLEXPORT},
+	{"errexit",ERREXIT},
+	{"norcs",NORCS},
+	{"histignorespace",HISTIGNORESPACE},
+	{"histignoredups",HISTIGNOREDUPS},
+	{"interactive",INTERACTIVE},
+	{"histlit",HISTLIT},
+	{"interactivecomments",INTERACTIVECOMMENTS},
+	{"login",LOGINSHELL},
+	{"monitor",MONITOR},
+	{"noexec",NOEXEC},
+	{"shinstdin",SHINSTDIN},
+	{"nounset",NOUNSET},
+	{"verbose",VERBOSE},
+	{"chaselinks",CHASELINKS},
+	{"xtrace",XTRACE},
+	{"shwordsplit",SHWORDSPLIT},
+	{"menucompletebeep",MENUCOMPLETEBEEP},
+	{"histnostore",HISTNOSTORE},
+	{"extendedglob",EXTENDEDGLOB},
+	{"globcomplete",GLOBCOMPLETE},
+	{"cshjunkiequotes",CSHJUNKIEQUOTES},
+	{"pushdminus",PUSHDMINUS},
+	{"cshjunkieloops",CSHJUNKIELOOPS},
+	{"rcquotes",RCQUOTES},
+	{"noshortloops",NOSHORTLOOPS},
+	{"lastmenu",LASTMENU},
+	{"automenu",AUTOMENU},
+	{"histverify",HISTVERIFY},
+	{"nolistbeep",NOLISTBEEP},
+	{"nohup",NOHUP},
+	{"noequals",NOEQUALS},
+	{"kshoptionprint",KSHOPTIONPRINT},
+	{"cshnullglob",CSHNULLGLOB},
+	{"hashcmds",HASHCMDS},
+	{"hashdirs",HASHDIRS},
+	{"numericglobsort",NUMERICGLOBSORT},
+	{"hashlistall",HASHLISTALL},
+	{"overstrike",OVERSTRIKE},
+	{"nohistbeep",NOHISTBEEP},
+	{"pushdignoredups",PUSHDIGNOREDUPS},
+	{"autoremoveslash",AUTOREMOVESLASH},
+	{"extendedhistory",EXTENDEDHISTORY},
+	{"appendhistory",APPENDHISTORY},
+	{NULL,0}
 };
 #endif
 
@@ -959,13 +1054,15 @@ struct option optns[] = {
 #define ALSTAT_JUNK 2	/* don't put word in history List */
 
 #undef isset
-#define isset(X) (opts[X])
-#define unset(X) (!opts[X])
+#define isset(X) (opts[(int)X] == OPT_SET)
+#define unset(X) (opts[(int)X] == OPT_UNSET)
 #define interact (isset(INTERACTIVE))
 #define jobbing (isset(MONITOR))
+#define jobbingv opts[MONITOR]
 #define nointr() signal(SIGINT,SIG_IGN)
 #define islogin (isset(LOGINSHELL))
 
+#ifndef SYSVR4
 #ifndef _IBMR2
 #undef WIFSTOPPED
 #undef WIFSIGNALED
@@ -973,7 +1070,7 @@ struct option optns[] = {
 #undef WEXITSTATUS
 #undef WTERMSIG
 #undef WSTOPSIG
-#undef WCOREDUMPED
+#undef WCOREDUMP
 
 #define WIFSTOPPED(X) (((X)&0377)==0177)
 #define WIFSIGNALED(X) (((X)&0377)!=0&&((X)&0377)!=0177)
@@ -982,7 +1079,8 @@ struct option optns[] = {
 #define WTERMSIG(X) ((X)&0177)
 #define WSTOPSIG(X) (((X)>>8)&0377)
 #endif
-#define WCOREDUMPED(X) ((X)&0200)
+#define WCOREDUMP(X) ((X)&0200)
+#endif
 
 #ifndef S_ISBLK
 #define	_IFMT		0170000
@@ -998,16 +1096,20 @@ struct option optns[] = {
 #define	S_ISREG(m)	(((m)&_IFMT) == _IFREG)
 #endif
 
-#ifndef S_ISSOCK
 #ifndef _IFMT
 #define _IFMT 0170000
 #endif
-#define	_IFLNK	0120000
+
+#ifndef S_ISSOCK
 #define	_IFSOCK	0140000
-#define	S_ISLNK(m)	(((m)&_IFMT) == _IFLNK)
 #define	S_ISSOCK(m)	(((m)&_IFMT) == _IFSOCK)
 #endif
 
+#ifndef S_ISLNK
+#define	_IFLNK	0120000
+#define	S_ISLNK(m)	(((m)&_IFMT) == _IFLNK)
+#endif
+
 #if S_IFIFO == S_IFSOCK
 #undef S_IFIFO
 #endif
@@ -1024,7 +1126,7 @@ EXTERN FILE *bshin;
 
 EXTERN char **path,**cdpath,**fpath,**watch,**mailpath;
 EXTERN char **manpath,**tildedirs,**fignore;
-EXTERN char **hosts;
+EXTERN char **hosts,**psvar;
 
 /* named directories */
 
@@ -1380,7 +1482,7 @@ EXTERN char *hptr;
 
 /* the current history line */
 
-EXTERN char *hline;
+EXTERN char *chline;
 
 /* the termcap buffer */
 
@@ -1398,7 +1500,7 @@ EXTERN int termok;
 
 EXTERN int badcshglob;
 
-/* max size of hline */
+/* max size of histline */
 
 EXTERN int hlinesz;
 
@@ -1431,6 +1533,14 @@ EXTERN Lklist heaplist;
 
 EXTERN int useheap;
 
+/* pid of process undergoing 'process substitution' */
+
+EXTERN int cmdoutpid;
+
+/* exit status of process undergoing 'process substitution' */
+
+EXTERN int cmdoutval;
+
 #include "signals.h"
 
 #ifdef GLOBALS
@@ -1439,7 +1549,7 @@ EXTERN int useheap;
 char **sigptr = sigs;
 
 /* tokens */
-char *ztokens = "#$^*()$=|{}[]`<>?~`,";
+char *ztokens = "#$^*()$=|{}[]`<>?~`,'\"\\";
 
 #else
 extern char *ztokens,**sigptr;
@@ -1460,7 +1570,7 @@ EXTERN List sigfuncs[VSIGCOUNT];
 
 /* $HISTCHARS */
 
-EXTERN char bangchar,hatchar,hashchar;
+EXTERN unsigned char bangchar,hatchar,hashchar;
 
 EXTERN int eofseen;
 
